from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from django.core.exceptions import ObjectDoesNotExist
from slack_sdk.models.blocks.block_elements import ButtonElement
from slack_sdk.models.blocks.blocks import ActionsBlock, Block, SectionBlock
from slack_sdk.models.views import View

from firefighter.confluence.models import PostMortemManager
from firefighter.incidents.enums import IncidentStatus
from firefighter.incidents.models.incident import Incident
from firefighter.slack.utils import respond
from firefighter.slack.views.modals.base_modal.base import SlackModal, app
from firefighter.slack.views.modals.base_modal.mixins import (
    IncidentSelectableModalMixin,
)

if TYPE_CHECKING:
    from slack_bolt.context.ack.ack import Ack


logger = logging.getLogger(__name__)


class PostMortemModal(
    IncidentSelectableModalMixin,
    SlackModal,
):
    open_action: str = "open_modal_create_postmortem"
    push_action: str = "push_modal_create_postmortem"
    callback_id: str = "incident_create_postmortem"

    def build_modal_fn(self, incident: Incident, **kwargs: Any) -> View:
        blocks: list[Block] = []

        # Check existing post-mortems
        has_confluence = _safe_has_relation(incident, "postmortem_for")
        has_jira = _safe_has_relation(incident, "jira_postmortem_for")

        if has_confluence or has_jira:
            blocks.append(
                SectionBlock(text=f"Post-mortem(s) for incident #{incident.id}:")
            )

            if has_confluence:
                blocks.append(
                    SectionBlock(
                        text=f"• Confluence: <{incident.postmortem_for.page_url}|View page>"
                    )
                )

            if has_jira:
                blocks.append(
                    SectionBlock(
                        text=f"• Jira: <{incident.jira_postmortem_for.issue_url}|{incident.jira_postmortem_for.jira_issue_key}>"
                    )
                )
        elif (
            incident.needs_postmortem
            and incident.status < IncidentStatus.MITIGATED
        ):
            blocks.append(
                SectionBlock(
                    text=f"Post-mortem for incident #{incident.id} will be automatically created when the incident reaches MITIGATED status."
                )
            )
        else:
            if incident.needs_postmortem:
                text = "Post-mortem was not automatically created for this incident. You can create it manually."
            else:
                text = "P3 incident post-mortem is not mandatory. You can still have one if you think is necessary by clicking on the button below."
            blocks.extend(
                [
                    SectionBlock(text=text),
                    ActionsBlock(
                        elements=[
                            ButtonElement(
                                text="Create post-mortem now",
                                action_id="incident_create_postmortem_now",
                                value=str(incident.id),
                            )
                        ]
                    ),
                ]
            )

        return View(
            type="modal",
            title="Postmortem"[:24],
            submit=None,
            callback_id=self.callback_id,
            private_metadata=str(incident.id),
            blocks=blocks,
        )

    @staticmethod
    def handle_modal_fn(ack: Ack, **_kwargs: Any) -> None:
        # This modal is now read-only (no submit button)
        # Post-mortems are created automatically when incident reaches MITIGATED status
        ack()


@app.action("incident_create_postmortem_now")
def handle_create_postmortem_action(ack: Ack, body: dict[str, Any]) -> None:
    """Create post-mortem(s) on demand from the modal (e.g. P3+ incidents)."""
    ack()

    incident_id = str(body.get("actions", [{}])[0].get("value", "")).strip()
    try:
        incident = Incident.objects.get(pk=incident_id)
    except Incident.DoesNotExist:
        respond(body, text=":x: Incident not found.")
        return

    if incident.status < IncidentStatus.MITIGATED:
        respond(
            body,
            text=f":x: Post-mortem can only be created when the incident is at least *Mitigated* (current status: *{incident.status.label}*).",
        )
        return

    try:
        confluence_pm, jira_pm = PostMortemManager.create_postmortem_for_incident(
            incident
        )
    except Exception:
        logger.exception("Failed to create post-mortem for incident #%s", incident_id)
        respond(body, text=":x: Failed to create post-mortem. Please try again.")
        return

    created_targets: list[str] = []
    if confluence_pm:
        created_targets.append("Confluence")
    if jira_pm:
        created_targets.append("Jira")

    if created_targets:
        targets = " and ".join(created_targets)
        respond(body, text=f":white_check_mark: {targets} post-mortem created.")
    else:
        respond(body, text=":warning: No post-mortem was created.")


modal_postmortem = PostMortemModal()


def _safe_has_relation(instance: Incident, attr: str) -> bool:
    """Safely check if a reverse relation exists without triggering KeyError in cache.

    Django's reverse OneToOne descriptor can raise KeyError when using hasattr
    on unsaved or freshly created instances. We guard against that here.
    """
    try:
        getattr(instance, attr)
    except (AttributeError, ObjectDoesNotExist, KeyError):
        return False
    else:
        return True



from __future__ import annotations

from django.urls import path

from firefighter.slack.views.views import SlackEventsHandler

app_name = "slack"
urlpatterns = [
    path("incident/", SlackEventsHandler.as_view(), name="slack_events_handler"),
]



from __future__ import annotations

from typing import Any
from unittest.mock import MagicMock, call

import pytest
from slack_sdk.errors import SlackApiError

from firefighter.slack.utils import respond


@pytest.fixture
def mock_slack_client() -> MagicMock:
    return MagicMock()


@pytest.fixture
def body_channel() -> dict[str, str]:
    return {
        "channel_id": "C12345",
        "user_id": "U12345",
    }


@pytest.fixture
def body_dm() -> dict[str, str]:
    return {"channel_name": "directmessage", "user_id": "U12345"}


@pytest.fixture
def body_no_user_id() -> dict[str, str | dict[str, str]]:
    return {"channel_id": "C12345", "user": {"id": "U12345"}}


@pytest.fixture
def body_no_channel_id() -> dict[str, str | dict[str, str]]:
    return {"channel": {"id": "C12345"}, "user_id": "U12345"}


@pytest.fixture
def body_view_submission() -> dict[str, str]:
    return {"type": "view_submission", "channel_id": "C12345", "user_id": "U12345"}


def test_respond_direct_message(mock_slack_client: MagicMock, body_dm: dict[str, str]):
    respond(body_dm, "test message", client=mock_slack_client)
    mock_slack_client.chat_postMessage.assert_called_once_with(
        channel="U12345", text="test message", blocks=None
    )


def test_respond_channel(mock_slack_client: MagicMock, body_channel: dict[str, str]):
    respond(body_channel, "test message", client=mock_slack_client)
    mock_slack_client.chat_postEphemeral.assert_called_once_with(
        user="U12345", channel="C12345", text="test message", blocks=None
    )


def test_respond_channel_fallback_to_dm(
    mock_slack_client: MagicMock, body_channel: dict[str, str]
):
    mock_slack_client.chat_postEphemeral.side_effect = SlackApiError(
        "error", MagicMock()
    )
    respond(body_channel, "test message", client=mock_slack_client)

    expected_calls = [
        call(
            channel="U12345",
            text=":warning: The bot could not respond in the channel you invoked it. Please add it to this channel or conversation if you want to interact with the bot there. If you believe this is a bug, please tell @pulse.",
        ),
        call(channel="U12345", text="test message", blocks=None),
    ]
    mock_slack_client.chat_postMessage.assert_has_calls(expected_calls)


@pytest.mark.parametrize("channel_name", ["directmessage", "channel"])
def test_respond_directmessage_and_channel(
    mock_slack_client: MagicMock,
    body_dm: dict[str, Any],
    body_channel: dict[str, Any],
    channel_name: str,
):
    body = body_dm if channel_name == "directmessage" else body_channel

    respond(body, "test message", client=mock_slack_client)

    if channel_name == "directmessage":
        mock_slack_client.chat_postMessage.assert_called_with(
            channel="U12345", text="test message", blocks=None
        )
    else:
        mock_slack_client.chat_postEphemeral.assert_called_with(
            user="U12345", channel="C12345", text="test message", blocks=None
        )


def test_respond_no_channel_id(
    mock_slack_client: MagicMock, body_no_channel_id: dict[str, Any]
):
    respond(body_no_channel_id, "test message", client=mock_slack_client)

    mock_slack_client.chat_postEphemeral.assert_called_with(
        user="U12345", channel="C12345", text="test message", blocks=None
    )


def test_respond_view_submission(
    mock_slack_client: MagicMock, body_view_submission: dict[str, Any]
):
    mock_slack_client.chat_postEphemeral.side_effect = SlackApiError(
        "error", MagicMock()
    )
    respond(body_view_submission, "test message", client=mock_slack_client)

    mock_slack_client.chat_postMessage.assert_called_once_with(
        channel="U12345", text="test message", blocks=None
    )


def test_no_user_id_no_channel_id(mock_slack_client: MagicMock):
    with pytest.raises(
        ValueError,
        match="At least one is required",
    ):
        respond({}, "test message", client=mock_slack_client)



from __future__ import annotations

from firefighter.jira_app.types import StatusTransitionInfo
from firefighter.jira_app.utils import (
    _states_to_transitions_names,
    _transition_path,
    get_status_id_from_name,
    get_transitions_to_apply,
)


def test__transition_path() -> None:
    statuses_transitions: dict[int, set[int]] = {1: {2, 3}, 2: {3}, 3: {4}, 4: set()}

    path = _transition_path(statuses_transitions, 1, 4)
    assert path == [1, 3, 4]

    path = _transition_path(statuses_transitions, 2, 4)
    assert path == [2, 3, 4]

    path = _transition_path(statuses_transitions, 3, 4)
    assert path == [3, 4]

    path = _transition_path(statuses_transitions, 1, 2)
    assert path == [1, 2]

    path = _transition_path(statuses_transitions, 4, 1)
    assert path == []


def test__states_to_transitions_names() -> None:
    statuses_transitions_names = {
        1: {2: "transition1", 3: "transition2"},
        2: {3: "transition3", 4: "transition4"},
        3: {4: "transition5"},
    }

    transitions = _states_to_transitions_names(statuses_transitions_names, [1, 2, 3, 4])
    assert transitions == ["transition1", "transition3", "transition5"]

    transitions = _states_to_transitions_names(statuses_transitions_names, [1, 3, 4])
    assert transitions == ["transition2", "transition5"]

    transitions = _states_to_transitions_names(statuses_transitions_names, [2, 3, 4])
    assert transitions == ["transition3", "transition5"]

    transitions = _states_to_transitions_names(statuses_transitions_names, [1, 2])
    assert transitions == ["transition1"]

    transitions = _states_to_transitions_names(statuses_transitions_names, [2, 4])
    assert transitions == ["transition4"]

    transitions = _states_to_transitions_names(statuses_transitions_names, [1])
    assert transitions == []


def test__get_transitions_to_apply() -> None:
    current_status_id = 1
    closed_state_id = 4
    transitions_info = [
        StatusTransitionInfo(
            status_id=1,
            status_name="Open",
            target_statuses={2, 3},
            transition_to_status={2: "Assign", 3: "Close"},
        ),
        StatusTransitionInfo(
            status_id=2,
            status_name="Assigned",
            target_statuses={3, 4},
            transition_to_status={3: "Reopen", 4: "Close"},
        ),
        StatusTransitionInfo(
            status_id=3,
            status_name="Closed",
            target_statuses={1},
            transition_to_status={1: "Reopen"},
        ),
        StatusTransitionInfo(
            status_id=4,
            status_name="Closed",
            target_statuses=set(),
            transition_to_status={},
        ),
    ]

    # Test if function returns correct transitions to close issue
    expected_transitions = ["Assign", "Close"]
    assert (
        get_transitions_to_apply(current_status_id, transitions_info, closed_state_id)
        == expected_transitions
    )

    # Test if function returns empty list if current_status_id is equal to closed_state_id
    current_status_id = 4
    assert (
        get_transitions_to_apply(current_status_id, transitions_info, closed_state_id)
        == []
    )

    # Test if function returns empty list if there's no path from current_status_id to closed_state_id
    current_status_id = 4
    closed_state_id = 1
    assert (
        get_transitions_to_apply(current_status_id, transitions_info, closed_state_id)
        == []
    )


def test__get_status_id_from_name() -> None:
    transitions_info = [
        StatusTransitionInfo(
            status_id=1,
            status_name="Open",
            target_statuses={2, 3},
            transition_to_status={2: "Assign", 3: "Close"},
        ),
        StatusTransitionInfo(
            status_id=2,
            status_name="Assigned",
            target_statuses={3, 4},
            transition_to_status={3: "Reopen", 4: "Close"},
        ),
        StatusTransitionInfo(
            status_id=3,
            status_name="Closed",
            target_statuses={1},
            transition_to_status={1: "Reopen"},
        ),
        StatusTransitionInfo(
            status_id=4,
            status_name="Resolved",
            target_statuses=set(),
            transition_to_status={},
        ),
    ]

    # Test if function returns correct status_id for a given status_name
    assert get_status_id_from_name(transitions_info, "Open") == 1
    assert get_status_id_from_name(transitions_info, "Assigned") == 2
    assert get_status_id_from_name(transitions_info, "Closed") == 3
    assert get_status_id_from_name(transitions_info, "Resolved") == 4

    # Test if function returns None for a non-existent status_name
    assert get_status_id_from_name(transitions_info, "NonExistent") is None



"""This module contains the logic to open an incident channel and invite responders.

XXX It might be divided into two signals/tasks, one for creating, one for inviting.
XXX Sending the end signal should be done by this signal's caller, not by this signal receiver directly.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from django.dispatch import receiver
from slack_sdk.errors import SlackApiError

from firefighter.incidents.signals import create_incident_conversation
from firefighter.slack.messages.slack_messages import (
    SlackMessageDeployWarning,
    SlackMessageIncidentDeclaredAnnouncement,
    SlackMessageIncidentDeclaredAnnouncementGeneral,
)
from firefighter.slack.models.conversation import Conversation
from firefighter.slack.models.incident_channel import IncidentChannel
from firefighter.slack.models.user import SlackUser
from firefighter.slack.rules import (
    should_publish_in_general_channel,
    should_publish_in_it_deploy_channel,
)
from firefighter.slack.signals import incident_channel_done

if TYPE_CHECKING:
    from firefighter.incidents.models.incident import Incident
    from firefighter.incidents.models.user import User

logger = logging.getLogger(__name__)


@receiver(signal=create_incident_conversation)
def create_incident_slack_conversation(
    incident: Incident,
    *_args: Any,
    **_kwargs: Any,
) -> int | None:
    """Main process to open an incident channel, set it up and invite responders. It MUST be called when an incident is created.

    Args:
        incident (Incident): The incident to open. It should be saved before calling this function, and have its first incident update created.

    Kwargs:
        jira_extra_fields (dict): Optional dictionary of customer/seller fields for Jira ticket

    """
    # Extract jira_extra_fields from kwargs
    jira_extra_fields = _kwargs.get("jira_extra_fields", {})
    channel: IncidentChannel | None = IncidentChannel.objects.create_incident_channel(
        incident=incident
    )
    if not channel:
        logger.warning("Giving up on Slack channel creation for %s.", incident.id)
        return None

    new_channel_id: str = channel.channel_id
    if not new_channel_id:
        logger.warning("Missing channel id for %s.", incident.id)
        return None
    if not (
        incident.created_by
        and hasattr(incident.created_by, "slack_user")
        and incident.created_by.slack_user
    ):
        if not hasattr(incident.created_by, "slack_user"):
            SlackUser.objects.add_slack_id_to_user(incident.created_by)
        logger.debug(incident.created_by)

    # Join the channel. We are already in the channel if it is a private channel.
    if not incident.private:
        channel.conversations_join()

    channel.set_incident_channel_topic()

    # Add the person that opened the incident in the channel
    if (
        incident.created_by
        and hasattr(incident.created_by, "slack_user")
        and incident.created_by.slack_user
        and incident.created_by.slack_user.slack_id
    ):
        try:
            channel.invite_users([incident.created_by])
        except SlackApiError:
            logger.warning(
                f"Could not import Slack opener user! Slack ID: {incident.created_by.slack_user.slack_id}, User {incident.created_by}, Channel ID {new_channel_id}",
                exc_info=True,
            )
    else:
        logger.warning("Could not find user Slack ID for opener_user!")

    # Send message in the created channel
    channel.send_message_and_save(
        SlackMessageIncidentDeclaredAnnouncement(incident), pin=True
    )

    # Post in general channel #tech-incidents if needed
    if should_publish_in_general_channel(incident, incident_update=None):
        announcement_general = SlackMessageIncidentDeclaredAnnouncementGeneral(incident)

        tech_incidents_conversation = Conversation.objects.get_or_none(
            tag="tech_incidents"
        )
        if tech_incidents_conversation:
            tech_incidents_conversation.send_message_and_save(announcement_general)
        else:
            logger.warning(
                "Could not find tech_incidents conversation! Is there a channel with tag tech_incidents?"
            )

    # Create a response team
    users_list: list[User] = incident.build_invite_list()

    # Invite all users
    incident.conversation.invite_users(users_list)

    # Post in #it-deploy if needed
    if should_publish_in_it_deploy_channel(incident):
        announcement_it_deploy = SlackMessageDeployWarning(incident)
        announcement_it_deploy.id = f"{announcement_it_deploy.id}_{incident.id}"

        it_deploy_conversation = Conversation.objects.get_or_none(tag="it_deploy")
        if it_deploy_conversation:
            it_deploy_conversation.send_message_and_save(announcement_it_deploy)
        else:
            logger.warning(
                "Could not find it_deploy conversation! Is there a channel with tag it_deploy?"
            )

    incident_channel_done.send_robust(
        sender=__name__,
        incident=incident,
        channel=channel,
        jira_extra_fields=jira_extra_fields,
    )
    return None



"""Django management command to backdate an incident's mitigated_at timestamp for testing."""

from __future__ import annotations

from datetime import timedelta
from typing import Any

from django.core.management.base import BaseCommand, CommandParser
from django.utils import timezone

from firefighter.incidents.models.incident import Incident


class Command(BaseCommand):
    """Backdate an incident's mitigated_at timestamp for testing post-mortem reminders."""

    help = "Backdate an incident's mitigated_at timestamp by a specified number of days"

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            "incident_id",
            type=int,
            help="ID of the incident to backdate",
        )
        parser.add_argument(
            "--days",
            type=int,
            default=6,
            help="Number of days to backdate (default: 6, to trigger 5-day reminder)",
        )
        parser.add_argument(
            "--reset",
            action="store_true",
            help="Reset mitigated_at to current time instead of backdating",
        )

    def handle(self, *args: Any, **options: Any) -> None:
        incident_id = options["incident_id"]
        days = options["days"]
        reset = options["reset"]

        try:
            incident = Incident.objects.get(id=incident_id)
        except Incident.DoesNotExist:
            self.stdout.write(
                self.style.ERROR(f"Incident #{incident_id} does not exist")
            )
            return

        if reset:
            incident.mitigated_at = timezone.now()
            incident.save(update_fields=["mitigated_at"])
            self.stdout.write(
                self.style.SUCCESS(
                    f"✅ Reset mitigated_at for incident #{incident_id} to current time: {incident.mitigated_at}"
                )
            )
        else:
            old_value = incident.mitigated_at
            new_value = timezone.now() - timedelta(days=days)
            incident.mitigated_at = new_value
            incident.save(update_fields=["mitigated_at"])

            self.stdout.write(
                self.style.SUCCESS(
                    f"✅ Backdated incident #{incident_id} mitigated_at:"
                )
            )
            self.stdout.write(f"   Old value: {old_value}")
            self.stdout.write(f"   New value: {new_value}")
            self.stdout.write(f"   Backdated by {days} days")

        self.stdout.write("\nIncident details:")
        self.stdout.write(f"   ID: {incident.id}")
        self.stdout.write(f"   Title: {incident.title}")
        self.stdout.write(f"   Priority: {incident.priority.name}")
        self.stdout.write(f"   Status: {incident.status.label}")
        self.stdout.write(f"   Needs postmortem: {incident.needs_postmortem}")
        self.stdout.write(f"   Mitigated at: {incident.mitigated_at}")

        if incident.needs_postmortem and days >= 5:
            self.stdout.write(
                self.style.WARNING(
                    "\n⚠️  This incident should now trigger a 5-day reminder!"
                )
            )
            self.stdout.write(
                "\nTo test the reminder, run:"
            )
            self.stdout.write(
                self.style.NOTICE(
                    "   pdm run python manage.py test_postmortem_reminders"
                )
            )



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from slack_sdk.errors import SlackApiError
from slack_sdk.models.blocks.basic_components import MarkdownTextObject, Option
from slack_sdk.models.blocks.block_elements import (
    PlainTextInputElement,
    StaticSelectElement,
)
from slack_sdk.models.blocks.blocks import (
    Block,
    ContextBlock,
    DividerBlock,
    InputBlock,
    SectionBlock,
)
from slack_sdk.models.views import View

from firefighter.incidents.enums import ClosureReason, IncidentStatus
from firefighter.slack.slack_templating import slack_block_footer, slack_block_separator
from firefighter.slack.utils import respond
from firefighter.slack.views.modals.base_modal.base import SlackModal
from firefighter.slack.views.modals.base_modal.mixins import (
    IncidentSelectableModalMixin,
)

if TYPE_CHECKING:
    from slack_bolt.context.ack.ack import Ack

    from firefighter.incidents.models.incident import Incident, User

logger = logging.getLogger(__name__)


class ClosureReasonModal(IncidentSelectableModalMixin, SlackModal):
    """Modal for closing an incident with a mandatory reason from early statuses."""

    open_action: str = "closure_reason_incident"
    open_shortcut = "closure_reason_incident"
    callback_id: str = "incident_closure_reason"

    def build_modal_fn(
        self, body: dict[str, Any], incident: Incident, **kwargs: Any  # noqa: ARG002
    ) -> View:
        """Build the closure reason modal."""
        # Build closure reason options (exclude RESOLVED)
        closure_options = [
            Option(
                value=choice[0],
                label=choice[1],
            )
            for choice in ClosureReason.choices
            if choice[0] != ClosureReason.RESOLVED
        ]

        blocks: list[Block] = [
            SectionBlock(
                text=f"*Closure Reason Required for Incident #{incident.id}*\n_{incident.title}_"
            ),
            ContextBlock(
                elements=[
                    MarkdownTextObject(
                        text=f"⚠️ This incident is currently in *{incident.status.label}* status.\nA closure reason is required to close incidents from this status."
                    )
                ]
            ),
            DividerBlock(),
            InputBlock(
                block_id="closure_reason",
                label="Closure Reason",
                element=StaticSelectElement(
                    action_id="select_closure_reason",
                    options=closure_options,
                    placeholder="Select a reason...",
                ),
                hint="Why is this incident being closed directly?",
            ),
            InputBlock(
                block_id="closure_reference",
                label="Reference (optional)",
                element=PlainTextInputElement(
                    action_id="input_closure_reference",
                    placeholder="e.g., #1234 or https://...",
                    max_length=100,
                ),
                hint="Related incident ID or external reference",
                optional=True,
            ),
            InputBlock(
                block_id="closure_message",
                label="Closure Message",
                element=PlainTextInputElement(
                    action_id="input_closure_message",
                    multiline=True,
                    placeholder="Brief explanation of why this incident is being closed...",
                ),
                hint="This message will be added to the incident timeline",
            ),
            slack_block_separator(),
            slack_block_footer(),
        ]

        return View(
            type="modal",
            title=f"Close #{incident.id}"[:24],
            submit="Close Incident",
            callback_id=self.callback_id,
            private_metadata=str(incident.id),
            blocks=blocks,
        )

    def handle_modal_fn(  # type: ignore[override]
        self, ack: Ack, body: dict[str, Any], incident: Incident, user: User
    ) -> bool | None:
        """Handle the closure reason modal submission."""
        # Extract form values up-front so validation can account for the submitted reason
        state_values = body["view"]["state"]["values"]
        closure_reason = state_values["closure_reason"]["select_closure_reason"][
            "selected_option"
        ]["value"]
        closure_reference = (
            state_values["closure_reference"]["input_closure_reference"].get(
                "value", ""
            )
            or ""
        )
        message = state_values["closure_message"]["input_closure_message"]["value"]

        # For early closure (OPEN/INVESTIGATING), we bypass normal workflow checks
        # For normal closure (MITIGATED/POST_MORTEM), we must validate key events
        current_status = incident.status
        is_early_closure = current_status.value in {IncidentStatus.OPEN, IncidentStatus.INVESTIGATING}

        if not is_early_closure:
            # Normal closure path - validate that incident can be closed
            can_close, reasons = incident.can_be_closed
            if not can_close:
                # Build error message from reasons
                error_messages = [reason[1] for reason in reasons]
                error_text = "\n".join([f"• {msg}" for msg in error_messages])
                ack(
                    response_action="errors",
                    errors={
                        "closure_message": f"Cannot close this incident:\n{error_text}"
                    }
                )
                return False
        else:
            # Early closure path - validate with the submitted closure reason
            # Temporarily inject the submitted closure_reason so early-closure bypass applies
            original_closure_reason = incident.closure_reason
            incident.closure_reason = closure_reason
            try:
                can_close, reasons = incident.can_be_closed
            finally:
                incident.closure_reason = original_closure_reason
            if not can_close:
                # Build error message from reasons
                error_messages = [reason[1] for reason in reasons]
                error_text = "\n".join([f"• {msg}" for msg in error_messages])
                ack(
                    response_action="errors",
                    errors={
                        "closure_message": f"Cannot close this incident:\n{error_text}"
                    },
                )
                return False

        # Clear ALL modals in the stack (not just this one)
        # This ensures the underlying "Update Status" modal is also closed
        ack(response_action="clear")

        try:
            # Update incident with closure fields
            incident.closure_reason = closure_reason
            incident.closure_reference = closure_reference
            incident.save(update_fields=["closure_reason", "closure_reference"])

            # Create incident update with closure
            incident.create_incident_update(
                created_by=user,
                status=IncidentStatus.CLOSED,
                message=message,
                event_type="closure_reason",
            )

        except Exception:
            logger.exception("Error closing incident #%s with reason", incident.id)
            respond(
                body=body,
                text=f"❌ Failed to close incident #{incident.id}",
            )
            return False
        else:
            # Send confirmation message
            try:
                respond(
                    body=body,
                    text=(
                        f"✅ Incident #{incident.id} has been closed.\n"
                        f"*Reason:* {ClosureReason(closure_reason).label}\n"
                        f"*Message:* {message}"
                        + (
                            f"\n*Reference:* {closure_reference}"
                            if closure_reference
                            else ""
                        )
                    ),
                )
            except SlackApiError as e:
                if e.response.get("error") == "messages_tab_disabled":
                    logger.warning(
                        "Cannot send DM to user %s - messages tab disabled",
                        user.email,
                    )
                else:
                    # Re-raise for other Slack API errors
                    raise

            logger.info(
                "Incident #%s closed with reason by %s: %s",
                incident.id,
                user.email,
                closure_reason,
            )
            return True

    def get_select_modal_title(self) -> str:
        return "Close with Reason"

    def get_select_title(self) -> str:
        return "Select an incident to close with a specific reason"


modal_closure_reason = ClosureReasonModal()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from django import forms

from firefighter.incidents.models.incident_membership import IncidentRole
from firefighter.incidents.models.incident_role_type import IncidentRoleType
from firefighter.incidents.models.user import User

logger = logging.getLogger(__name__)
if TYPE_CHECKING:
    from collections.abc import Iterable

    from django.forms import Field

    from firefighter.incidents.models.incident import Incident


class IncidentUpdateRolesForm(forms.Form):
    incident: Incident
    user: User | None
    initial: dict[str, Any]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.incident: Incident = kwargs.pop("incident")
        self.user = kwargs.pop("user", None)
        super().__init__(*args, **kwargs)
        self.generate_fields_dynamically()

    def generate_fields_dynamically(self) -> None:
        roles_types_base = IncidentRoleType.objects.order_by("order")
        self.incident_roles = self.get_incident_roles(roles_types_base)

        for incident_role in self.incident_roles:
            field_name = f"role_{incident_role.role_type.slug}"
            self.fields[field_name] = forms.ModelChoiceField(
                queryset=User.objects.all(),
                required=incident_role.role_type.required,
                label=incident_role.role_type.name,
                help_text=(
                    f"{incident_role.role_type.emoji} "
                    if incident_role.role_type.emoji
                    else ""
                )
                + incident_role.role_type.summary,
                initial=incident_role.user if hasattr(incident_role, "user") else None,
            )

            logger.debug(
                f"Initial {field_name} was set to {self.initial.get(field_name)}"
            )

    def clean(self) -> dict[str, Any] | None:
        if self.user is None:
            raise forms.ValidationError("User is required")
        # XXX Validate that we have all fields and no extra fields
        return super().clean()

    def save(self) -> None:
        """Custom save method to save the updated roles."""
        self.fields: dict[str, Field]
        roles_to_update: dict[str, User | None] = {}
        for field_name in self.fields:
            if field_name not in self.changed_data or self.fields[field_name].disabled:
                logger.debug("skipping %s", field_name)
                continue
            if not field_name.startswith("role_"):
                continue

            role_type_slug = field_name.replace("role_", "")

            roles_to_update[role_type_slug] = self.cleaned_data[field_name]
        if self.user is None:
            raise RuntimeError("User is required, was the form validated?")
        self.incident.update_roles(self.user, roles_to_update)

    def get_incident_roles(
        self,
        incident_role_types: Iterable[IncidentRoleType],
    ) -> list[IncidentRole]:
        incident_roles: list[IncidentRole] = []
        for incident_role_type in incident_role_types:
            # Get the IncidentRole if it exists in DB, or create it locally with User=None
            try:
                incident_role = IncidentRole.objects.get(
                    incident=self.incident, role_type=incident_role_type
                )
            except IncidentRole.DoesNotExist:
                incident_role = IncidentRole(
                    incident=self.incident, role_type=incident_role_type, user=None
                )
            incident_roles.append(incident_role)
        return incident_roles



from __future__ import annotations

import operator
from datetime import datetime

from django.utils import timezone
from factory import Iterator, LazyAttribute, Sequence, SubFactory
from factory.django import DjangoModelFactory
from factory.faker import Faker
from factory.fuzzy import FuzzyChoice, FuzzyDateTime, FuzzyInteger

from firefighter.incidents.enums import IncidentStatus
from firefighter.incidents.models import (
    Environment,
    Group,
    Incident,
    IncidentCategory,
    Priority,
    User,
)


class UserFactory(DjangoModelFactory[User]):
    class Meta:
        model = User

    name = Faker("name")  # type: ignore[no-untyped-call]
    email = Faker("email")  # type: ignore[no-untyped-call]
    username = Sequence(lambda n: f"user_{n}")  # type: ignore[no-untyped-call]


class GroupFactory(DjangoModelFactory[Group]):
    class Meta:
        model = Group

    name = Faker("text", max_nb_chars=20)  # type: ignore[no-untyped-call]
    description = Faker("text", max_nb_chars=50)  # type: ignore[no-untyped-call]
    order = FuzzyInteger(100, 1000)  # type: ignore[no-untyped-call]


class PriorityFactory(DjangoModelFactory[Priority]):
    """Factory for creating Priority instances in tests.

    Creates Priority objects with random values. Use specific values in tests
    when testing priority-specific behavior (e.g., P1-P5 JIRA mapping).
    """
    class Meta:
        model = Priority

    name = Faker("text", max_nb_chars=20)  # type: ignore[no-untyped-call]
    value = FuzzyInteger(0, 100)  # type: ignore[no-untyped-call]
    order = FuzzyInteger(100, 1000)  # type: ignore[no-untyped-call]


# Legacy alias - will be removed when Severity model is removed
SeverityFactory = PriorityFactory


class EnvironmentFactory(DjangoModelFactory[Environment]):
    class Meta:
        model = Environment

    name = Faker("text", max_nb_chars=20)  # type: ignore[no-untyped-call]
    description = Faker("text", max_nb_chars=50)  # type: ignore[no-untyped-call]
    order = FuzzyInteger(100, 1000)  # type: ignore[no-untyped-call]


class IncidentCategoryFactory(DjangoModelFactory[IncidentCategory]):
    class Meta:
        model = IncidentCategory

    name = Faker("text", max_nb_chars=30)  # type: ignore[no-untyped-call]
    description = Faker("text", max_nb_chars=50)  # type: ignore[no-untyped-call]
    order = FuzzyInteger(100, 1000)  # type: ignore[no-untyped-call]
    group = SubFactory(GroupFactory)  # type: ignore[no-untyped-call]


class IncidentFactory(DjangoModelFactory[Incident]):
    class Meta:
        model = Incident

    id = Sequence(lambda n: n + 1)  # type: ignore[no-untyped-call]
    title = Faker("text", max_nb_chars=50)  # type: ignore[no-untyped-call]
    description = Faker("text")  # type: ignore[no-untyped-call]
    _status = FuzzyChoice(IncidentStatus.choices, getter=operator.itemgetter(0))  # type: ignore[no-untyped-call]
    created_at = FuzzyDateTime(
        start_dt=datetime(2018, 1, 1, tzinfo=timezone.get_current_timezone()),
        end_dt=datetime(2022, 1, 1, tzinfo=timezone.get_current_timezone()),
    )  # type: ignore[no-untyped-call]
    updated_at = LazyAttribute(
        lambda inc: datetime(
            inc.created_at.year,
            inc.created_at.month,
            inc.created_at.day,
            tzinfo=timezone.get_current_timezone(),
        )
    )  # type: ignore[no-untyped-call]
    incident_category = Iterator(IncidentCategory.objects.all())  # type: ignore[no-untyped-call]
    priority = Iterator(Priority.objects.all())  # type: ignore[no-untyped-call]
    environment = Iterator(Environment.objects.all())  # type: ignore[no-untyped-call]

    created_by = SubFactory(UserFactory)  # type: ignore[no-untyped-call]



# Generated by Django 4.2.23 on 2025-09-17 17:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("incidents", "0026_alter_incidentcategory_options_and_more"),
        ("slack", "0007_remove_components_fields"),
    ]

    operations = [
        migrations.AlterField(
            model_name="conversation",
            name="incident_categories",
            field=models.ManyToManyField(
                blank=True,
                related_name="conversations",
                to="incidents.incidentcategory",
            ),
        ),
        migrations.AlterField(
            model_name="usergroup",
            name="incident_categories",
            field=models.ManyToManyField(
                blank=True,
                help_text="Incident created with this usergroup automatically add the group members to these issue categories.",
                related_name="usergroups",
                to="incidents.incidentcategory",
            ),
        ),
    ]



"""End-to-end integration tests for unified incident form custom fields propagation.

This test file verifies that custom fields (zendesk_ticket_id, seller_contract_id, etc.)
are properly transmitted from the form submission all the way to:
1. Form cleaned_data
2. Jira extra fields in signals
3. Incident custom_fields
"""
from __future__ import annotations

import inspect
from typing import Any
from unittest.mock import MagicMock, patch

import pytest

from firefighter.incidents.forms.unified_incident import UnifiedIncidentForm
from firefighter.incidents.models.impact import ImpactLevel, ImpactType, LevelChoices
from firefighter.incidents.signals import create_incident_conversation
from firefighter.slack.messages.base import SlackMessageStrategy
from firefighter.slack.views.modals.open import OpenModal
from firefighter.slack.views.modals.opening.details.unified import (
    OpeningUnifiedModal,
    UnifiedIncidentFormSlack,
)


def create_mock_slack_message(incident):
    """Create a mock SlackMessageIncidentDeclaredAnnouncement with proper attributes."""
    mock_slack_message_instance = MagicMock()
    mock_slack_message_instance.strategy = SlackMessageStrategy.APPEND
    mock_slack_message_instance.id = "test_message"
    mock_slack_message_instance.incident = incident  # Store the real incident
    mock_slack_message_instance.incident_update = None  # Declarations don't have incident updates
    mock_slack_message_instance.get_slack_message_params.return_value = {
        "text": "Test incident declared",
        "blocks": []
    }
    return mock_slack_message_instance


def setup_jira_mocks(mock_select_impact_used, mock_select_impact_source,
                     mock_get_jira_user, mock_jira_client, mock_prepare_fields,
                     mock_jira_ticket_model, mock_slack_announcement):
    """Configure all Jira-related mocks for the unified workflow."""
    # Mock SelectImpactForm - we need to configure BOTH patches
    mock_form_instance = MagicMock()
    mock_form_instance.save.return_value = None
    mock_form_instance.is_valid.return_value = True
    mock_select_impact_used.return_value = mock_form_instance
    mock_select_impact_source.return_value = mock_form_instance

    # Mock Jira user lookup
    mock_jira_user = MagicMock()
    mock_jira_user.id = "jira-user-123"
    mock_get_jira_user.return_value = mock_jira_user

    # Mock prepare_jira_fields
    mock_prepare_fields.return_value = {"summary": "Test", "project": {"key": "TEST"}}

    # Mock Jira API create_issue call
    mock_jira_client.create_issue.return_value = {
        "issue_key": "TEST-123",
        "issue_url": "https://jira.example.com/browse/TEST-123",
    }

    # Mock JiraTicket creation
    mock_jira_ticket = MagicMock()
    mock_jira_ticket.url = "https://jira.example.com/browse/TEST-123"
    mock_jira_ticket.id = "TEST-123"
    mock_jira_ticket_model.objects.create.return_value = mock_jira_ticket

    # Mock SlackMessageIncidentDeclaredAnnouncement
    mock_slack_announcement.side_effect = create_mock_slack_message


@pytest.mark.django_db
class TestUnifiedIncidentFormCustomFieldsPropagation:
    """Test end-to-end propagation of custom fields from form to signal."""

    def test_customer_impact_zendesk_field_propagates_to_signal(
        self,
        priority_factory,
        environment_factory,
        incident_category_factory,
        user_factory,
    ):
        """Zendesk ticket ID should propagate from form to signal with jira_extra_fields."""
        # Setup
        priority = priority_factory(value=1, default=True)
        environment = environment_factory(value="PRD", default=True)
        category = incident_category_factory()

        # Get customer impact from fixtures (impact_type=3, value=HI)
        customers_impact_type = ImpactType.objects.get(value="customers_impact")
        customer_impact = ImpactLevel.objects.get(
            impact_type=customers_impact_type,
            value=LevelChoices.HIGH.value
        )

        # Form data with customer impact and zendesk ticket
        form_data = {
            "title": "Customer issue with Zendesk ticket",
            "description": "Test incident with customer impact and Zendesk tracking",
            "incident_category": category.id,
            "environment": [environment.id],
            "platform": ["platform-FR"],
            "priority": priority.id,
            "zendesk_ticket_id": "12345",
        }

        impacts_data = {
            "customers_impact": customer_impact,
        }

        # Create form with impacts context
        form = UnifiedIncidentFormSlack(
            data=form_data,
            impacts_data=impacts_data,
            response_type="critical",
        )

        # Verify form is valid
        assert form.is_valid(), f"Form should be valid. Errors: {form.errors}"

        # Verify zendesk_ticket_id is in cleaned_data
        assert "zendesk_ticket_id" in form.cleaned_data
        assert form.cleaned_data["zendesk_ticket_id"] == "12345"

        # Mock signal receiver to capture jira_extra_fields
        signal_received = False
        signal_data: dict[str, Any] = {}

        def capture_signal(sender: Any, incident: Any, **kwargs: Any) -> None:
            nonlocal signal_received, signal_data
            signal_received = True
            signal_data = {"incident": incident, **kwargs}

        # Connect signal receiver
        create_incident_conversation.connect(capture_signal, weak=False)

        try:
            # Mock Jira-related calls in unified workflow
            # SelectImpactForm appears in TWO places and needs TWO patches:
            # 1. Module-level import (line 11) → patch where it's USED: unified_incident.SelectImpactForm
            # 2. Lazy import inside _create_jira_ticket (line 414) → patch SOURCE: select_impact.SelectImpactForm
            # Other Jira imports are lazy, so we patch them at source modules
            with patch("firefighter.incidents.forms.unified_incident.SelectImpactForm") as mock_select_impact_used, \
                 patch("firefighter.incidents.forms.select_impact.SelectImpactForm") as mock_select_impact_source, \
                 patch("firefighter.raid.service.get_jira_user_from_user") as mock_get_jira_user, \
                 patch("firefighter.raid.client.client") as mock_jira_client, \
                 patch("firefighter.raid.forms.prepare_jira_fields") as mock_prepare_fields, \
                 patch("firefighter.raid.forms.set_jira_ticket_watchers_raid"), \
                 patch("firefighter.raid.models.JiraTicket") as mock_jira_ticket_model, \
                 patch("firefighter.slack.messages.slack_messages.SlackMessageIncidentDeclaredAnnouncement") as mock_slack_announcement:

                setup_jira_mocks(mock_select_impact_used, mock_select_impact_source,
                                mock_get_jira_user, mock_jira_client, mock_prepare_fields,
                                mock_jira_ticket_model, mock_slack_announcement)

                # Trigger workflow
                user = user_factory()
                form.trigger_incident_workflow(
                    creator=user,
                    impacts_data=impacts_data,
                    response_type="critical",
                )

            # Verify signal was sent
            assert signal_received, "Signal create_incident_conversation should have been sent"

            # Verify jira_extra_fields was passed to signal
            assert "jira_extra_fields" in signal_data, "jira_extra_fields should be in signal kwargs"

            # Verify zendesk_ticket_id is in jira_extra_fields
            jira_extra_fields = signal_data["jira_extra_fields"]
            assert "zendesk_ticket_id" in jira_extra_fields
            assert jira_extra_fields["zendesk_ticket_id"] == "12345"

            # Verify incident was created with custom_fields
            incident = signal_data["incident"]
            assert incident is not None
            assert hasattr(incident, "custom_fields")
            assert "zendesk_ticket_id" in incident.custom_fields
            assert incident.custom_fields["zendesk_ticket_id"] == "12345"

        finally:
            # Disconnect signal
            create_incident_conversation.disconnect(capture_signal)

    def test_seller_impact_fields_propagate_to_signal(
        self,
        priority_factory,
        environment_factory,
        incident_category_factory,
        user_factory,
    ):
        """Seller-specific fields should propagate from form to signal with jira_extra_fields."""
        # Setup
        priority = priority_factory(value=1, default=True)
        environment = environment_factory(value="PRD", default=True)
        category = incident_category_factory()

        # Get seller impact from fixtures (impact_type=2, value=HI)
        sellers_impact_type = ImpactType.objects.get(value="sellers_impact")
        seller_impact = ImpactLevel.objects.get(
            impact_type=sellers_impact_type,
            value=LevelChoices.HIGH.value
        )

        # Form data with seller impact and seller-specific fields
        form_data = {
            "title": "Seller issue with contract tracking",
            "description": "Test incident with seller impact and contract ID",
            "incident_category": category.id,
            "environment": [environment.id],
            "platform": ["platform-FR"],
            "priority": priority.id,
            "seller_contract_id": "CONTRACT-789",
            "zoho_desk_ticket_id": "ZOHO-456",
            "is_key_account": True,
            "is_seller_in_golden_list": False,
        }

        impacts_data = {
            "sellers_impact": seller_impact,
        }

        # Create form with impacts context
        form = UnifiedIncidentFormSlack(
            data=form_data,
            impacts_data=impacts_data,
            response_type="critical",
        )

        # Verify form is valid
        assert form.is_valid(), f"Form should be valid. Errors: {form.errors}"

        # Verify seller fields are in cleaned_data
        assert form.cleaned_data["seller_contract_id"] == "CONTRACT-789"
        assert form.cleaned_data["zoho_desk_ticket_id"] == "ZOHO-456"
        assert form.cleaned_data["is_key_account"] is True
        assert form.cleaned_data["is_seller_in_golden_list"] is False

        # Mock signal receiver to capture jira_extra_fields
        signal_received = False
        signal_data: dict[str, Any] = {}

        def capture_signal(sender: Any, incident: Any, **kwargs: Any) -> None:
            nonlocal signal_received, signal_data
            signal_received = True
            signal_data = {"incident": incident, **kwargs}

        # Connect signal receiver
        create_incident_conversation.connect(capture_signal, weak=False)

        try:
            # Mock Jira-related calls in unified workflow
            # SelectImpactForm appears in TWO places and needs TWO patches:
            # 1. Module-level import (line 11) → patch where it's USED: unified_incident.SelectImpactForm
            # 2. Lazy import inside _create_jira_ticket (line 414) → patch SOURCE: select_impact.SelectImpactForm
            # Other Jira imports are lazy, so we patch them at source modules
            with patch("firefighter.incidents.forms.unified_incident.SelectImpactForm") as mock_select_impact_used, \
                 patch("firefighter.incidents.forms.select_impact.SelectImpactForm") as mock_select_impact_source, \
                 patch("firefighter.raid.service.get_jira_user_from_user") as mock_get_jira_user, \
                 patch("firefighter.raid.client.client") as mock_jira_client, \
                 patch("firefighter.raid.forms.prepare_jira_fields") as mock_prepare_fields, \
                 patch("firefighter.raid.forms.set_jira_ticket_watchers_raid"), \
                 patch("firefighter.raid.models.JiraTicket") as mock_jira_ticket_model, \
                 patch("firefighter.slack.messages.slack_messages.SlackMessageIncidentDeclaredAnnouncement") as mock_slack_announcement:

                setup_jira_mocks(mock_select_impact_used, mock_select_impact_source,
                                mock_get_jira_user, mock_jira_client, mock_prepare_fields,
                                mock_jira_ticket_model, mock_slack_announcement)

                # Trigger workflow
                user = user_factory()
                form.trigger_incident_workflow(
                    creator=user,
                    impacts_data=impacts_data,
                    response_type="critical",
                )

            # Verify signal was sent
            assert signal_received, "Signal create_incident_conversation should have been sent"

            # Verify jira_extra_fields was passed to signal
            assert "jira_extra_fields" in signal_data

            # Verify all seller fields are in jira_extra_fields
            jira_extra_fields = signal_data["jira_extra_fields"]
            assert jira_extra_fields["seller_contract_id"] == "CONTRACT-789"
            assert jira_extra_fields["zoho_desk_ticket_id"] == "ZOHO-456"
            assert jira_extra_fields["is_key_account"] is True
            assert jira_extra_fields["is_seller_in_golden_list"] is False

            # Verify incident was created with all seller fields in custom_fields
            incident = signal_data["incident"]
            assert "seller_contract_id" in incident.custom_fields
            assert incident.custom_fields["seller_contract_id"] == "CONTRACT-789"
            assert "zoho_desk_ticket_id" in incident.custom_fields
            assert incident.custom_fields["zoho_desk_ticket_id"] == "ZOHO-456"
            assert "is_key_account" in incident.custom_fields
            assert incident.custom_fields["is_key_account"] is True
            assert "is_seller_in_golden_list" in incident.custom_fields
            assert incident.custom_fields["is_seller_in_golden_list"] is False

        finally:
            # Disconnect signal
            create_incident_conversation.disconnect(capture_signal)

    def test_both_customer_and_seller_fields_propagate_together(
        self,
        priority_factory,
        environment_factory,
        incident_category_factory,
        user_factory,
    ):
        """Both customer and seller fields should propagate when both impacts are selected."""
        # Setup
        priority = priority_factory(value=1, default=True)
        environment = environment_factory(value="PRD", default=True)
        category = incident_category_factory()

        # Get impacts from fixtures
        customers_impact_type = ImpactType.objects.get(value="customers_impact")
        customer_impact = ImpactLevel.objects.get(
            impact_type=customers_impact_type,
            value=LevelChoices.HIGH.value
        )

        sellers_impact_type = ImpactType.objects.get(value="sellers_impact")
        seller_impact = ImpactLevel.objects.get(
            impact_type=sellers_impact_type,
            value=LevelChoices.LOW.value
        )

        # Form data with both impacts and all custom fields
        form_data = {
            "title": "Combined customer and seller issue",
            "description": "Test incident affecting both customers and sellers",
            "incident_category": category.id,
            "environment": [environment.id],
            "platform": ["platform-All"],
            "priority": priority.id,
            "zendesk_ticket_id": "ZD-999",
            "seller_contract_id": "SC-888",
            "zoho_desk_ticket_id": "ZOHO-777",
            "is_key_account": False,
            "is_seller_in_golden_list": True,
        }

        impacts_data = {
            "customers_impact": customer_impact,
            "sellers_impact": seller_impact,
        }

        # Create form with impacts context
        form = UnifiedIncidentFormSlack(
            data=form_data,
            impacts_data=impacts_data,
            response_type="critical",
        )

        # Verify form is valid
        assert form.is_valid(), f"Form should be valid. Errors: {form.errors}"

        # Mock signal receiver
        signal_received = False
        signal_data: dict[str, Any] = {}

        def capture_signal(sender: Any, incident: Any, **kwargs: Any) -> None:
            nonlocal signal_received, signal_data
            signal_received = True
            signal_data = {"incident": incident, **kwargs}

        create_incident_conversation.connect(capture_signal, weak=False)

        try:
            # Mock Jira-related calls in unified workflow
            # SelectImpactForm appears in TWO places and needs TWO patches:
            # 1. Module-level import (line 11) → patch where it's USED: unified_incident.SelectImpactForm
            # 2. Lazy import inside _create_jira_ticket (line 414) → patch SOURCE: select_impact.SelectImpactForm
            # Other Jira imports are lazy, so we patch them at source modules
            with patch("firefighter.incidents.forms.unified_incident.SelectImpactForm") as mock_select_impact_used, \
                 patch("firefighter.incidents.forms.select_impact.SelectImpactForm") as mock_select_impact_source, \
                 patch("firefighter.raid.service.get_jira_user_from_user") as mock_get_jira_user, \
                 patch("firefighter.raid.client.client") as mock_jira_client, \
                 patch("firefighter.raid.forms.prepare_jira_fields") as mock_prepare_fields, \
                 patch("firefighter.raid.forms.set_jira_ticket_watchers_raid"), \
                 patch("firefighter.raid.models.JiraTicket") as mock_jira_ticket_model, \
                 patch("firefighter.slack.messages.slack_messages.SlackMessageIncidentDeclaredAnnouncement") as mock_slack_announcement:

                setup_jira_mocks(mock_select_impact_used, mock_select_impact_source,
                                mock_get_jira_user, mock_jira_client, mock_prepare_fields,
                                mock_jira_ticket_model, mock_slack_announcement)

                # Trigger workflow
                user = user_factory()
                form.trigger_incident_workflow(
                    creator=user,
                    impacts_data=impacts_data,
                    response_type="critical",
                )

            # Verify signal was sent with all fields
            assert signal_received
            jira_extra_fields = signal_data["jira_extra_fields"]

            # Verify both customer and seller fields are present
            assert jira_extra_fields["zendesk_ticket_id"] == "ZD-999"
            assert jira_extra_fields["seller_contract_id"] == "SC-888"
            assert jira_extra_fields["zoho_desk_ticket_id"] == "ZOHO-777"
            assert jira_extra_fields["is_key_account"] is False
            assert jira_extra_fields["is_seller_in_golden_list"] is True

            # Verify new fields: environments and platforms
            assert "environments" in jira_extra_fields
            assert jira_extra_fields["environments"] == ["PRD"]
            assert "platforms" in jira_extra_fields
            assert jira_extra_fields["platforms"] == ["platform-All"]

            # Verify incident custom_fields has all fields (5 original + 2 new = 7)
            incident = signal_data["incident"]
            assert len(incident.custom_fields) == 7
            assert incident.custom_fields["zendesk_ticket_id"] == "ZD-999"
            assert incident.custom_fields["seller_contract_id"] == "SC-888"
            assert incident.custom_fields["environments"] == ["PRD"]
            assert incident.custom_fields["platforms"] == ["platform-All"]

        finally:
            create_incident_conversation.disconnect(capture_signal)

    def test_no_impact_sends_empty_jira_extra_fields(
        self,
        priority_factory,
        environment_factory,
        incident_category_factory,
        user_factory,
    ):
        """When no customer/seller impact is selected, jira_extra_fields should only contain None values."""
        # Setup
        priority = priority_factory(value=1, default=True)
        environment = environment_factory(value="PRD", default=True)
        category = incident_category_factory()

        # Form data without any custom fields
        form_data = {
            "title": "Basic incident without impacts",
            "description": "Test incident with no customer or seller impact",
            "incident_category": category.id,
            "environment": [environment.id],
            "platform": ["platform-FR"],
            "priority": priority.id,
        }

        impacts_data: dict[str, Any] = {}

        # Create form without impacts context
        form = UnifiedIncidentFormSlack(
            data=form_data,
            impacts_data=impacts_data,
            response_type="critical",
        )

        # Verify form is valid
        assert form.is_valid(), f"Form should be valid. Errors: {form.errors}"

        # Verify custom fields are NOT in form fields (removed by _configure_field_visibility)
        assert "zendesk_ticket_id" not in form.fields
        assert "seller_contract_id" not in form.fields

        # Mock signal receiver
        signal_received = False
        signal_data: dict[str, Any] = {}

        def capture_signal(sender: Any, incident: Any, **kwargs: Any) -> None:
            nonlocal signal_received, signal_data
            signal_received = True
            signal_data = {"incident": incident, **kwargs}

        create_incident_conversation.connect(capture_signal, weak=False)

        try:
            # Mock Jira-related calls in unified workflow
            # SelectImpactForm appears in TWO places and needs TWO patches:
            # 1. Module-level import (line 11) → patch where it's USED: unified_incident.SelectImpactForm
            # 2. Lazy import inside _create_jira_ticket (line 414) → patch SOURCE: select_impact.SelectImpactForm
            # Other Jira imports are lazy, so we patch them at source modules
            with patch("firefighter.incidents.forms.unified_incident.SelectImpactForm") as mock_select_impact_used, \
                 patch("firefighter.incidents.forms.select_impact.SelectImpactForm") as mock_select_impact_source, \
                 patch("firefighter.raid.service.get_jira_user_from_user") as mock_get_jira_user, \
                 patch("firefighter.raid.client.client") as mock_jira_client, \
                 patch("firefighter.raid.forms.prepare_jira_fields") as mock_prepare_fields, \
                 patch("firefighter.raid.forms.set_jira_ticket_watchers_raid"), \
                 patch("firefighter.raid.models.JiraTicket") as mock_jira_ticket_model, \
                 patch("firefighter.slack.messages.slack_messages.SlackMessageIncidentDeclaredAnnouncement") as mock_slack_announcement:

                setup_jira_mocks(mock_select_impact_used, mock_select_impact_source,
                                mock_get_jira_user, mock_jira_client, mock_prepare_fields,
                                mock_jira_ticket_model, mock_slack_announcement)

                # Trigger workflow
                user = user_factory()
                form.trigger_incident_workflow(
                    creator=user,
                    impacts_data=impacts_data,
                    response_type="critical",
                )

            # Verify signal was sent
            assert signal_received

            # Verify jira_extra_fields exists - custom fields are None, but environments/platforms have defaults
            jira_extra_fields = signal_data["jira_extra_fields"]
            assert jira_extra_fields["zendesk_ticket_id"] is None
            assert jira_extra_fields["seller_contract_id"] is None
            assert jira_extra_fields["zoho_desk_ticket_id"] is None
            assert jira_extra_fields["is_key_account"] is None
            assert jira_extra_fields["is_seller_in_golden_list"] is None

            # New fields: environments and platforms have default values (not None)
            assert "environments" in jira_extra_fields
            assert jira_extra_fields["environments"] == ["PRD"]  # Default from form
            assert "platforms" in jira_extra_fields
            assert jira_extra_fields["platforms"] == ["platform-FR"]  # Default from form

            # Verify incident custom_fields only has environments and platforms (None values filtered out)
            incident = signal_data["incident"]
            assert len(incident.custom_fields) == 2
            assert incident.custom_fields["environments"] == ["PRD"]
            assert incident.custom_fields["platforms"] == ["platform-FR"]

        finally:
            create_incident_conversation.disconnect(capture_signal)


@pytest.mark.django_db
class TestOpenModalPreservesCustomFieldsContext:
    """Test that open.py properly passes context to form during validation and submission."""

    def test_open_modal_passes_impacts_data_to_form_validation(
        self,
        priority_factory,
        environment_factory,
        incident_category_factory,
    ):
        """OpenModal._validate_details_form should pass open_incident_context to form initialization."""
        # Setup
        priority = priority_factory(value=1, default=True)
        environment_factory(value="PRD", default=True)
        incident_category_factory()

        # Get customer impact from fixtures
        customers_impact_type = ImpactType.objects.get(value="customers_impact")
        customer_impact = ImpactLevel.objects.get(
            impact_type=customers_impact_type,
            value=LevelChoices.HIGH.value
        )

        # Simulate the context as it exists in open.py
        open_incident_context = {
            "response_type": "critical",
            "impact_form_data": {
                "customers_impact": customer_impact,
            },
            "details_form_data": {
                "priority": str(priority.id),
            },
        }

        # Mock form data with zendesk field
        details_form_data = {
            "title": "Test incident",
            "description": "Test description for validation",
            "zendesk_ticket_id": "VALIDATION-123",
        }

        # Call _validate_details_form with context (this is the fix we applied)
        _is_valid, _form_class, form = OpenModal._validate_details_form(
            details_form_modal_class=OpeningUnifiedModal,
            details_form_data=details_form_data,
            open_incident_context=open_incident_context,
        )

        # Verify that the form has zendesk_ticket_id field
        # This proves that impacts_data was passed correctly from open_incident_context
        assert "zendesk_ticket_id" in form.fields, (
            "zendesk_ticket_id should be in form fields - "
            "this proves open_incident_context was passed correctly"
        )

        # Form will be invalid due to missing required fields, but that's expected
        # We're only testing that the context was passed and fields were configured correctly

    def test_open_modal_handle_modal_fn_passes_context_to_form_submission(
        self,
        priority_factory,
        environment_factory,
        incident_category_factory,
        user_factory,
    ):
        """OpenModal.handle_modal_fn should pass impacts_data and response_type to form during submission."""
        # This test verifies that the fix in open.py lines 612-657 works correctly
        # Setup
        priority = priority_factory(value=1, default=True)
        environment = environment_factory(value="PRD", default=True)
        category = incident_category_factory()

        # Get customer impact from fixtures
        customers_impact_type = ImpactType.objects.get(value="customers_impact")
        customer_impact = ImpactLevel.objects.get(
            impact_type=customers_impact_type,
            value=LevelChoices.HIGH.value
        )

        # Simulate the context as OpenModal.handle_modal_fn would have it
        impacts_data_for_context = {
            "customers_impact": customer_impact,
        }

        form_data = {
            "title": "Submission test with Zendesk",
            "description": "Test that form receives context during submission",
            "incident_category": category.id,
            "environment": [environment.id],
            "platform": ["platform-FR"],
            "priority": priority.id,
            "zendesk_ticket_id": "SUBMIT-456",
        }

        # This simulates what happens in open.py after our fix (using inspect.signature)
        init_params = inspect.signature(UnifiedIncidentForm.__init__).parameters

        form_kwargs: dict[str, Any] = {}
        if "impacts_data" in init_params:
            form_kwargs["impacts_data"] = impacts_data_for_context
        if "response_type" in init_params:
            form_kwargs["response_type"] = "critical"

        # Create form with the kwargs (this is what open.py does now)
        form = UnifiedIncidentForm(form_data, **form_kwargs)

        # Verify form is valid
        assert form.is_valid(), f"Form should be valid. Errors: {form.errors}"

        # Verify zendesk_ticket_id is in cleaned_data
        assert form.cleaned_data["zendesk_ticket_id"] == "SUBMIT-456"

        # Verify that trigger_incident_workflow will send the field to signal
        signal_received = False
        signal_data: dict[str, Any] = {}

        def capture_signal(sender: Any, incident: Any, **kwargs: Any) -> None:
            nonlocal signal_received, signal_data
            signal_received = True
            signal_data = {"incident": incident, **kwargs}

        create_incident_conversation.connect(capture_signal, weak=False)

        try:
            # Mock Jira-related calls in unified workflow
            # SelectImpactForm appears in TWO places and needs TWO patches:
            # 1. Module-level import (line 11) → patch where it's USED: unified_incident.SelectImpactForm
            # 2. Lazy import inside _create_jira_ticket (line 414) → patch SOURCE: select_impact.SelectImpactForm
            # Other Jira imports are lazy, so we patch them at source modules
            with patch("firefighter.incidents.forms.unified_incident.SelectImpactForm") as mock_select_impact_used, \
                 patch("firefighter.incidents.forms.select_impact.SelectImpactForm") as mock_select_impact_source, \
                 patch("firefighter.raid.service.get_jira_user_from_user") as mock_get_jira_user, \
                 patch("firefighter.raid.client.client") as mock_jira_client, \
                 patch("firefighter.raid.forms.prepare_jira_fields") as mock_prepare_fields, \
                 patch("firefighter.raid.forms.set_jira_ticket_watchers_raid"), \
                 patch("firefighter.raid.models.JiraTicket") as mock_jira_ticket_model, \
                 patch("firefighter.slack.messages.slack_messages.SlackMessageIncidentDeclaredAnnouncement") as mock_slack_announcement:

                # Mock SelectImpactForm - we need to configure BOTH patches
                # They both return the same mock instance
                mock_form_instance = MagicMock()
                mock_form_instance.save.return_value = None
                mock_form_instance.is_valid.return_value = True
                mock_select_impact_used.return_value = mock_form_instance
                mock_select_impact_source.return_value = mock_form_instance

                # Mock Jira user lookup
                mock_jira_user = MagicMock()
                mock_jira_user.id = "jira-user-123"
                mock_get_jira_user.return_value = mock_jira_user

                # Mock prepare_jira_fields to return valid fields
                mock_prepare_fields.return_value = {"summary": "Test", "project": {"key": "TEST"}}

                # Mock Jira API create_issue call
                mock_jira_client.create_issue.return_value = {
                    "issue_key": "TEST-123",
                    "issue_url": "https://jira.example.com/browse/TEST-123",
                }

                # Mock JiraTicket creation - must have .url attribute for bookmarks
                mock_jira_ticket = MagicMock()
                mock_jira_ticket.url = "https://jira.example.com/browse/TEST-123"
                mock_jira_ticket.id = "TEST-123"
                mock_jira_ticket_model.objects.create.return_value = mock_jira_ticket

                # Mock SlackMessageIncidentDeclaredAnnouncement
                mock_slack_announcement.side_effect = create_mock_slack_message

                user = user_factory()
                form.trigger_incident_workflow(
                    creator=user,
                    impacts_data=impacts_data_for_context,
                    response_type="critical",
                )

            assert signal_received
            assert signal_data["jira_extra_fields"]["zendesk_ticket_id"] == "SUBMIT-456"

        finally:
            create_incident_conversation.disconnect(capture_signal)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from django.dispatch.dispatcher import receiver

from firefighter.incidents.enums import IncidentStatus
from firefighter.incidents.signals import incident_key_events_updated, incident_updated
from firefighter.slack.messages.slack_messages import (
    SlackMessageDeployWarning,
    SlackMessageIncidentDowngradeHint,
    SlackMessageIncidentRolesUpdated,
    SlackMessageIncidentStatusUpdated,
)
from firefighter.slack.models.conversation import Conversation, ConversationType
from firefighter.slack.rules import (
    should_publish_in_general_channel,
    should_publish_in_it_deploy_channel,
)

if TYPE_CHECKING:
    from django.db.models.fields.related_descriptors import ManyRelatedManager

    from firefighter.incidents.models.incident import Incident
    from firefighter.incidents.models.incident_update import IncidentUpdate
    from firefighter.incidents.models.priority import Priority
    from firefighter.incidents.models.user import User
logger = logging.getLogger(__name__)
# pylint: disable=unused-argument


@receiver(signal=incident_updated, sender="update_status")
def incident_updated_update_status_handler(
    sender: Any,
    incident: Incident,
    incident_update: IncidentUpdate,
    updated_fields: list[str],
    **kwargs: Any,
) -> None:
    # Skip Slack operations for incidents without channels (e.g., P4-P5)
    if not hasattr(incident, "conversation"):
        logger.debug(f"Skipping Slack channel update for incident {incident.id} (no conversation)")
        return

    # Update Slack channel if needed
    incident.conversation.rename_if_needed()

    # Update topic if needed
    if (
        "priority_id" in updated_fields
        or "incident_category_id" in updated_fields
        or "_status" in updated_fields
    ):
        incident.conversation.set_incident_channel_topic()

    publish_status_update(
        incident=incident,
        incident_update=incident_update,
        status_changed=bool("_status" in updated_fields),
        old_priority=kwargs.get("old_priority"),
    )


@receiver(signal=incident_updated, sender="update_status")
def incident_updated_check_dowmgrade_handler(
    sender: Any,
    incident: Incident,
    incident_update: IncidentUpdate,
    updated_fields: list[str],
    **kwargs: Any,
) -> None:
    # Skip Slack operations for incidents without channels (e.g., P4-P5)
    if not hasattr(incident, "conversation"):
        return

    # Only show downgrade hint when downgrading from critical (P1/P2/P3) to normal (P4/P5)
    old_priority: Priority | None = kwargs.get("old_priority")

    # Check if priority was actually changed
    if not incident_update.priority or "priority_id" not in updated_fields:
        return

    # New priority must be P4 or P5 (normal incident)
    new_priority_is_normal = incident_update.priority.value >= 4

    # Old priority must have been P1, P2, or P3 (critical incident)
    old_priority_was_critical = old_priority is not None and old_priority.value <= 3

    # Only show hint if downgrading from critical to normal
    if not (new_priority_is_normal and old_priority_was_critical):
        return

    try:
        commander_role = (
            incident.roles_set.filter(role_type__slug="commander")
            .select_related("user__slack_user")
            .get()
        )
        slack_user = commander_role.user.slack_user
    except incident.roles_set.model.DoesNotExist:
        slack_user = None

    if slack_user is None:
        if (
            incident_update.created_by is None
            or not hasattr(incident_update.created_by, "slack_user")
            or incident_update.created_by.slack_user is None
        ):
            return
        slack_user = incident_update.created_by.slack_user

    incident.conversation.send_message_ephemeral(
        message=SlackMessageIncidentDowngradeHint(
            incident=incident, incident_update=incident_update
        ),
        user=slack_user,
    )


@receiver(signal=incident_updated, sender="update_roles")
def incident_updated_update_roles_handler(
    sender: Any,
    incident: Incident,
    incident_update: IncidentUpdate,
    updated_fields: list[str],
    **kwargs: Any,
) -> None:
    # Skip Slack operations for incidents without channels (e.g., P4-P5)
    if not hasattr(incident, "conversation"):
        return

    # Publish a message to the conversation
    update_roles_message = SlackMessageIncidentRolesUpdated(
        incident=incident,
        incident_update=incident_update,
        first_update=False,
        updated_fields=updated_fields,
    )
    incident.conversation.send_message_and_save(update_roles_message)

    # Check if updated_fields map to existing IncidentRoleTypes that have a IncidentRole for this incident
    # If so, invite the users
    users_with_roles = incident.roles_set.filter(
        role_type__slug__in=[
            updated_field.removesuffix("_id") for updated_field in updated_fields
        ]
    ).select_related("user", "user__slack_user")
    users_to_invite = [
        role.user for role in users_with_roles if role.user.slack_user is not None
    ]

    incident.conversation.invite_users(list(users_to_invite))


@receiver(signal=incident_updated)
# pylint: disable=unused-argument
def incident_updated_reinvite_handler(
    sender: Any,
    incident: Incident,
    incident_update: IncidentUpdate,
    updated_fields: list[str],
    **kwargs: Any,
) -> None:
    # Skip Slack operations for incidents without channels (e.g., P4-P5)
    if not hasattr(incident, "conversation"):
        return

    if incident.conversation.type == ConversationType.PRIVATE_CHANNEL:
        return

    # Rebuild the invite list
    full_invite_list: set[User] = set(incident.build_invite_list())

    # Check if users from the invite list are not in the channel
    members_manager: ManyRelatedManager[User] = incident.conversation.members  # type: ignore [assignment]
    already_invited_users = members_manager.all()
    missing_users = full_invite_list.difference(already_invited_users)

    if len(missing_users) > 0:
        incident.conversation.invite_users(list(missing_users))
    else:
        logger.info(f"No users to re-invite for incident {incident.id}")


@receiver(signal=incident_key_events_updated)
def incident_key_events_updated_handler(
    sender: Any,
    incident: Incident,
    **kwargs: Any,
) -> None:
    logger.info("Received incident_key_events_updated signal")

    # Skip Slack operations for incidents without channels (e.g., P4-P5)
    if not hasattr(incident, "conversation"):
        logger.debug(f"Skipping key events update for incident {incident.id} (no conversation)")
        return

    # Skip key events for incidents closed directly
    if incident.closure_reason:
        logger.info(f"Skipping key events update for incident {incident.id} (direct closure)")
        return

    from firefighter.slack.views.modals.key_event_message import (
        SlackMessageKeyEvents,
    )

    incident.conversation.send_message_and_save(SlackMessageKeyEvents(incident))


def publish_status_update(
    incident: Incident,
    incident_update: IncidentUpdate,
    *,
    status_changed: bool = False,
    old_priority: Priority | None = None,
) -> None:
    """Publishes an update to the incident status."""
    # Skip Slack operations for incidents without channels (e.g., P4-P5)
    if not hasattr(incident, "conversation"):
        logger.debug(f"Skipping status update publication for incident {incident.id} (no conversation)")
        return

    message = SlackMessageIncidentStatusUpdated(
        incident=incident,
        incident_update=incident_update,
        in_channel=True,
    )
    incident.conversation.send_message_and_save(message)

    # Post to #tech-incidents
    if should_publish_in_general_channel(
        incident=incident, incident_update=incident_update, old_priority=old_priority
    ):
        publish_update_in_general_channel(
            incident=incident,
            incident_update=incident_update,
            status_changed=status_changed,
            old_priority=old_priority,
        )

    if (
        incident.ask_for_milestones
        and status_changed
        and incident.status >= IncidentStatus.MITIGATED
        and not incident.closure_reason  # Don't show key events for direct closures
    ):
        from firefighter.slack.views.modals.key_event_message import (
            SlackMessageKeyEvents,
        )

        incident.conversation.send_message_and_save(
            SlackMessageKeyEvents(incident=incident)
        )

    if should_publish_in_it_deploy_channel(incident=incident):
        announcement_it_deploy = SlackMessageDeployWarning(incident)
        announcement_it_deploy.id = f"{announcement_it_deploy.id}_{incident.id}"

        it_deploy_conversation = Conversation.objects.get_or_none(tag="it_deploy")
        if it_deploy_conversation:
            it_deploy_conversation.send_message_and_save(announcement_it_deploy)
        else:
            logger.warning(
                "Could not find it_deploy conversation! Is there a channel with tag it_deploy?"
            )


def publish_update_in_general_channel(
    incident: Incident,
    incident_update: IncidentUpdate,
    *,
    status_changed: bool = False,
    old_priority: Priority | None = None,
) -> None:
    update_status_message_global = SlackMessageIncidentStatusUpdated(
        incident_update=incident_update,
        incident=incident,
        in_channel=False,
        status_changed=status_changed,
        old_priority=old_priority,
    )
    tech_incidents_conversation = Conversation.objects.get_or_none(tag="tech_incidents")
    if tech_incidents_conversation:
        tech_incidents_conversation.send_message_and_save(update_status_message_global)
    else:
        logger.warning(
            "Could not find tech_incidents conversation! Is there a channel with tag tech_incidents?"
        )



from __future__ import annotations

from firefighter.incidents.models.environment import Environment
from firefighter.incidents.models.group import Group
from firefighter.incidents.models.impact import Impact
from firefighter.incidents.models.incident import Incident
from firefighter.incidents.models.incident_category import IncidentCategory
from firefighter.incidents.models.incident_cost import IncidentCost
from firefighter.incidents.models.incident_cost_type import IncidentCostType
from firefighter.incidents.models.incident_role_type import IncidentRoleType
from firefighter.incidents.models.incident_update import IncidentUpdate
from firefighter.incidents.models.milestone_type import MilestoneType
from firefighter.incidents.models.priority import Priority
from firefighter.incidents.models.severity import Severity
from firefighter.incidents.models.user import User



from __future__ import annotations

from django.conf import settings
from django.urls import URLPattern, URLResolver, include, path
from drf_spectacular.views import SpectacularAPIView, SpectacularSwaggerView
from rest_framework import routers

from firefighter.api import views

app_name = "api"

# Routers provide an easy way of automatically determining the URL conf.
router = routers.DefaultRouter()
router.include_root_view = False

router.register(
    r"environments",
    views.environments.EnvironmentViewSet,
    basename="environments",
)
router.register(
    r"incident_cost_types",
    views.incident_cost_types.IncidentCostTypeViewSet,
    basename="incident_cost_types",
)
router.register(
    r"incident_costs",
    views.incident_costs.IncidentCostViewSet,
    basename="incident_costs",
)
router.register(
    r"severities",
    views.severities.PriorityViewSet,
    basename="priorities",
)
# Legacy endpoint for backward compatibility
router.register(
    r"components",
    views.components.IncidentCategoryViewSet,
    basename="components",
)
# New preferred endpoint
router.register(
    r"incident-categories",
    views.components.IncidentCategoryViewSet,
    basename="incident-categories",
)
router.register(
    r"groups",
    views.groups.GroupViewSet,
    basename="groups",
)
router.register(
    r"incidents",
    views.incidents.IncidentViewSet,
    basename="incidents",
)


urlpatterns: list[URLPattern | URLResolver] = [
    path("", include(router.urls)),
    path(
        "incidents",
        views.incidents.CreateIncidentViewSet.as_view({"post": "create"}),  # pyright: ignore[reportGeneralTypeIssues]
        name="incidents",
    ),
]
if settings.FF_EXPOSE_API_DOCS:
    urlpatterns.extend((
        path(
            "schema",
            SpectacularAPIView.as_view(),  # pyright: ignore[reportGeneralTypeIssues]
            name="schema",
        ),
        path(
            "schema/swagger-ui",
            SpectacularSwaggerView.as_view(url_name="api:schema"),  # pyright: ignore[reportGeneralTypeIssues]
            name="swagger-ui",
        ),
    ))



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, cast

from django_filters.views import FilterView
from django_tables2.views import SingleTableMixin

from firefighter.confluence.models import Runbook, RunbookFilterSet
from firefighter.confluence.tables import RunbookTable

if TYPE_CHECKING:
    from firefighter.firefighter.utils import HtmxHttpRequest

logger = logging.getLogger(__name__)


class RunbooksViewList(SingleTableMixin, FilterView):
    model = Runbook
    queryset = Runbook.objects.all()
    context_object_name = "runbooks"

    filterset_class = RunbookFilterSet

    table_class = RunbookTable
    paginate_by = 150
    paginate_orphans = 20

    def get_template_names(self) -> list[str]:
        request = cast("HtmxHttpRequest", self.request)
        if request.htmx and not request.htmx.boosted:
            template_name = "layouts/partials/partial_table_list_paginated.html"
        else:
            template_name = "pages/runbook_list.html"

        return [template_name]

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        """No *args to pass."""
        context = super().get_context_data(**kwargs)
        context["filter_order"] = [
            "search",
            "service_type",
        ]
        context["page_title"] = "Runbooks List"
        return context



# tests/test_fields.py
from __future__ import annotations

import pytest
from django.forms import Form

from firefighter.incidents.enums import IncidentStatus
from firefighter.incidents.forms.utils import EnumChoiceField, GroupedModelChoiceField
from firefighter.incidents.models.group import Group
from firefighter.incidents.models.incident_category import IncidentCategory


class EnumChoiceFieldForm(Form):
    status = EnumChoiceField(enum_class=IncidentStatus)


class GroupedModelChoiceFieldForm(Form):
    incident_category = GroupedModelChoiceField(
        choices_groupby="group", queryset=IncidentCategory.objects.all()
    )


@pytest.fixture
def group() -> Group:
    return Group.objects.create(name="Group 1")


@pytest.fixture
def incident_categories(group: Group):
    return [
        IncidentCategory.objects.create(name="Issue category 1", group=group, order=1),
        IncidentCategory.objects.create(name="Issue category 2", group=group, order=2),
    ]


def test_enum_choice_field_valid() -> None:
    form = EnumChoiceFieldForm({"status": IncidentStatus.OPEN})

    assert form.is_valid()
    assert form.cleaned_data["status"] == IncidentStatus.OPEN


def test_enum_choice_field_invalid() -> None:
    form = EnumChoiceFieldForm({"status": 999})
    assert not form.is_valid()
    assert "status" in form.errors


@pytest.mark.django_db
def test_grouped_model_choice_field_valid(incident_categories: list[IncidentCategory]):
    form = GroupedModelChoiceFieldForm({"incident_category": incident_categories[0].id})
    assert form.is_valid()
    assert form.cleaned_data["incident_category"] == incident_categories[0]


def test_grouped_model_choice_field_invalid() -> None:
    form = GroupedModelChoiceFieldForm({"incident_category": "non-existent-id"})
    assert not form.is_valid()
    assert "incident_category" in form.errors


@pytest.fixture(scope="module")
def test_grouped_model_choice_field_grouping(
    incident_categories: list[IncidentCategory],
):
    form = GroupedModelChoiceFieldForm()
    grouped_choices = list(
        form.fields["incident_category"].iterator(field=form.fields["incident_category"])
    )
    assert len(grouped_choices) == 2  # One for the empty choice, and one for the group
    assert grouped_choices[0] == ("", form.fields["incident_category"].empty_label)
    assert grouped_choices[1][0] == incident_categories[0].group
    assert len(grouped_choices[1][1]) == 2  # Two components in the group



from __future__ import annotations

import logging
from typing import Any

from celery import shared_task

from firefighter.slack.models import SlackUser

logger = logging.getLogger(__name__)


@shared_task(
    name="slack.update_users_from_slack",
    retry_kwargs={"max_retries": 2},
    default_retry_delay=90,
)
def update_users_from_slack(*_args: Any, **_options: Any) -> None:
    """Retrieves users from Slack and updates the database (e.g. name, new profile picture, email, active status...)."""
    queryset = SlackUser.objects.all().order_by("user__updated_at")[:100]
    for slack_user in queryset:
        slack_user.update_user_info()



from __future__ import annotations

import logging
import uuid
from typing import TYPE_CHECKING, Any, ClassVar

from django.contrib import admin
from django.contrib.auth.models import AbstractUser
from django.contrib.auth.models import UserManager as BaseUserManager
from django.db import models
from django.urls import reverse
from django_stubs_ext.db.models import TypedModelMeta

if TYPE_CHECKING:
    from django.apps import apps
    from django_stubs_ext.db.models.manager import RelatedManager

    from firefighter.incidents.models.incident import IncidentRole

    if apps.is_installed("firefighter.slack"):
        from firefighter.slack.models.user import SlackUser
    if apps.is_installed("firefighter.pagerduty"):
        from firefighter.pagerduty.models import PagerDutyUser
    if apps.is_installed("firefighter.jira_app"):
        from firefighter.jira_app.models import JiraUser


logger = logging.getLogger(__name__)


class UserManager(BaseUserManager["User"]):
    def get_or_none(self, **kwargs: Any) -> User | None:
        try:
            return self.get(**kwargs)
        except User.DoesNotExist:
            return None


class User(AbstractUser):
    objects: ClassVar[UserManager] = UserManager()

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(
        max_length=128,
        help_text="User fullname. Prefer to set `first_name` and `last_name`, and get `full_name`. ",
    )
    email = models.EmailField(max_length=128, unique=True)
    bot = models.BooleanField(
        default=False,
        help_text="Used for non-person accounts (e.g. integrations that have bot/shared accounts.",
    )

    avatar = models.URLField(
        max_length=1024,
        blank=True,
        null=True,
        help_text="URL to user avatar. Can be hosted externally.",
    )

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta(TypedModelMeta):
        ordering = ["username"]

    def __str__(self) -> str:
        return self.full_name

    @property
    @admin.display(ordering="first_name")
    def full_name(self) -> str:
        """User full name (first + last name). Looks for the first_name and last_name fields, then name, then username. Will return an empty string if none of these are set."""
        if self.first_name and self.last_name:
            return f"{self.first_name} {self.last_name}"
        if self.name:
            return self.name
        if self.username:
            return self.username
        return ""

    def get_absolute_url(self) -> str:
        return reverse("incidents:user-detail", kwargs={"user_id": self.id})

    if TYPE_CHECKING:
        roles_set: RelatedManager[IncidentRole]
        if apps.is_installed("firefighter.slack"):
            slack_user: SlackUser | None
            slack_user_id: uuid.UUID | None
        if apps.is_installed("firefighter.pagerduty"):
            pagerduty_user: PagerDutyUser | None
            pagerduty_user_id: uuid.UUID | None
        if apps.is_installed("firefighter.raid"):
            jira_user: JiraUser | None
            jira_user_id: str | None



from __future__ import annotations

from firefighter.api.serializers import IncidentCostSerializer
from firefighter.api.views._base import ReadOnlyModelViewSet
from firefighter.incidents.models.incident_cost import IncidentCost


class IncidentCostViewSet(ReadOnlyModelViewSet[IncidentCost]):
    queryset = IncidentCost.objects.all().select_related("cost_type")
    serializer_class = IncidentCostSerializer



from __future__ import annotations

from typing import TYPE_CHECKING, Never

from django.db.models import Prefetch, QuerySet
from drf_spectacular.utils import OpenApiExample, OpenApiParameter, extend_schema
from rest_framework import mixins, status, viewsets
from rest_framework.renderers import JSONRenderer
from rest_framework.response import Response

from firefighter.api.serializers import IncidentSerializer
from firefighter.api.views._base import AdvancedGenericViewSet
from firefighter.incidents.models.incident import Incident, IncidentFilterSet
from firefighter.incidents.models.incident_cost import IncidentCost
from firefighter.incidents.models.incident_membership import IncidentRole
from firefighter.incidents.models.metric_type import IncidentMetric
from firefighter.incidents.signals import create_incident_conversation

if TYPE_CHECKING:
    from rest_framework.request import Request
    from rest_framework.serializers import BaseSerializer


class ProcessAfterResponse(Response):
    """Custom DRF Response, to trigger the Slack workflow after creating the incident and returning HTTP 201."""

    def close(self) -> None:
        super().close()
        incident = Incident.objects.get(id=self.data["id"])
        create_incident_conversation.send(
            "api",
            incident=incident,
        )


@extend_schema(
    parameters=[
        OpenApiParameter(
            name="fields",
            type=str,
            location=OpenApiParameter.QUERY,
            description="Comma-separated list of fields to include in a CSV or TSV render. Nested objects are de-nested and accessible with a dot `.`. Non-existent fields will generate empty rows.",
            default="__all__",
            examples=[
                OpenApiExample(
                    name="Comma separated list of fields",
                    value="id,name,incident_category.name,incident_category.group.name",
                    request_only=True,
                ),
                OpenApiExample(
                    name="All fields (default)",
                    value="__all__",
                    request_only=True,
                ),
            ],
        ),
        OpenApiParameter(
            name="show_tags",
            type=bool,
            location=OpenApiParameter.QUERY,
            description="Whether to include tags in the output.",
            default=False,
            examples=[OpenApiExample(name="Show tags", value=True, request_only=True)],
        ),
    ],
)
class IncidentViewSet(
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    AdvancedGenericViewSet[Incident],
):
    queryset = (
        Incident.objects.all()
        .select_related(
            "priority",
            "incident_category__group",
            "incident_category",
            "environment",
            "conversation",
            "created_by",
            "jira_ticket",
        )
        .prefetch_related(
            Prefetch(
                "metric_set",
                queryset=IncidentMetric.objects.all()
                .order_by("id")
                .select_related("metric_type"),
                to_attr="metrics_prefetched",
            ),
            Prefetch(
                "incident_cost_set",
                queryset=IncidentCost.objects.all()
                .order_by("id")
                .select_related("cost_type"),
                to_attr="costs_prefetched",
            ),
            Prefetch(
                "roles_set",
                queryset=IncidentRole.objects.all()
                .order_by("id")
                .select_related("role_type", "user"),
                to_attr="roles_prefetched",
            ),
        )
    )
    serializer_class = IncidentSerializer
    serializer_context = {"remove_fields": ["tags"]}
    filterset_class = IncidentFilterSet
    fields = [
        "id",
        "status",
        "environment.value",
        "priority.name",
        "title",
        "description",
        "incident_category.name",
        "incident_category.group.name",
        "created_at",
        "slack_channel_name",
        "status_page_url",
        "jira_ticket_key",
        "jira_ticket_url",
        "metrics.*.duration_seconds",
        "costs.*.amount",
        "roles.*.email",
    ]

    def list(self, request: Request, *args: Never, **kwargs: Never) -> Response:
        """List all incidents. Allows to show or hide tags which is DB intensive, with the show_tags parameters.

        `status` and `_status__*` uses an enum:
        ```
        10, "Open"
        20, "Investigating"
        30, "Mitigating"
        40, "Mitigated"
        50, "Post-mortem"
        60, "Closed"
        ```
        """
        queryset = self.filter_queryset(self.get_queryset())
        serializer: BaseSerializer[Incident]
        if request.query_params.get("show_tags") == "true":
            serializer = IncidentSerializer(
                queryset, many=True, context={"remove_fields": []}
            )
        else:
            serializer = self.get_serializer(
                queryset, many=True, context={"remove_fields": ["tags"]}
            )
        return Response(serializer.data)


@extend_schema(
    examples=[
        OpenApiExample(
            "Create an incident",
            summary="Create an incident",
            description="Create an incident, on INT, P4, with `Other` incident category and John Doe as creator. All fields are required. The email must be a valid email of a ManoMano employee, that has a Slack account or has already used FireFighter before.",
            value={
                "title": "Title of the incident, limited to 128 characters.",
                "description": "Longer description of the incident. No characters limit.",
                "environment_id": "1b960430-995b-47e1-beab-23dbe3dbccbf",
                "incident_category_id": "390a993a-d273-4db8-b7d6-190ab294961a",
                "priority_id": "b814c9d2-48a8-4ac4-9c71-ff844e1b77f1",
                "created_by_email": "john.doe@mycompany.com",
            },
            request_only=True,  # signal that example only applies to requests
            response_only=False,  # signal that example only applies to responses
        ),
        OpenApiExample(
            "Create incident response",
            status_codes=["201"],
            summary="Create incident response",
            description="Successfully created incident. Notice that the Slack channel is not created yet. You can retrieve the Slack channel by making a GET call after a few seconds.",
            value={
                "id": 886,
                "title": "Title of the incident, limited to 128 characters.",
                "status": "Open",
                "description": "Longer description of the incident. No characters limit.",
                "created_at": "2022-02-18T12:16:10.960407+01:00",
                "environment": {
                    "id": "1b960430-995b-47e1-beab-23dbe3dbccbf",
                    "value": "INT",
                    "name": "Int",
                    "description": "Integration environment",
                    "order": 3,
                    "default": False,
                },
                "incident_category": {
                    "id": "390a993a-d273-4db8-b7d6-190ab294961a",
                    "name": "Other",
                    "description": "",
                    "order": 200,
                    "created_at": "2019-12-13T15:44:53+01:00",
                    "updated_at": "2021-07-30T15:28:52.692000+02:00",
                    "group": {
                        "id": "13a1b0b2-9ff9-4c2b-b665-f497102459d7",
                        "name": "Other",
                        "description": "",
                        "order": 100000,
                        "created_at": "2020-04-14T18:12:42+02:00",
                        "updated_at": "2021-08-02T11:57:09.306000+02:00",
                    },
                },
                "priority": {
                    "id": "b814c9d2-48a8-4ac4-9c71-ff844e1b77f1",
                    "name": "P4",
                    "value": 4,
                    "emoji": "🌦️",
                    "description": "Minor issue not affecting customers.",
                    "order": 3,
                    "default": True,
                    "needs_postmortem": False,
                    "created_at": "2021-02-19T17:54:03.586000+01:00",
                    "updated_at": "2021-06-25T18:00:28.819000+02:00",
                },
                "slack_channel_name": "#None",
                "status_page_url": "https://incidents.mycompany.com/incident/1234",
                "tags": [],
                "metrics": [],
                "costs": [],
                "created_by": {
                    "name": "John Doe",
                    "email": "john.doe@mycompany.com",
                },
                "commander": {
                    "name": "John Doe",
                    "email": "john.doe@mycompany.com",
                },
                "communication_lead": {
                    "name": "John Doe",
                    "email": "john.doe@mycompany.com",
                },
            },
            request_only=False,
            response_only=True,
        ),
    ]
)
class CreateIncidentViewSet(
    mixins.CreateModelMixin,
    viewsets.GenericViewSet[Incident],
):
    queryset: QuerySet[Incident] = Incident.objects.all().select_related(
        "priority", "incident_category__group", "incident_category", "environment", "conversation"
    )
    serializer_class = IncidentSerializer
    filterset_class = IncidentFilterSet
    renderer_classes = [JSONRenderer]

    def create(self, request: Request, *args: Never, **kwargs: Never) -> Response:
        """Allow to create an incident.
        Requires a valid Bearer token, that you can create in the back-office if you have the right permissions.
        """
        serializer: BaseSerializer[Incident] = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        return ProcessAfterResponse(
            serializer.data, status=status.HTTP_201_CREATED, headers=headers
        )

    def perform_create(self, serializer: BaseSerializer[Incident]) -> None:
        serializer.save()



from __future__ import annotations

from types import SimpleNamespace

from firefighter.incidents.enums import IncidentStatus
from firefighter.slack.views.modals.postmortem import PostMortemModal


class TestPostMortemModal:
    @staticmethod
    def _get_text_blocks(view_dict: dict) -> list[str]:
        return [
            block.get("text", {}).get("text", "")
            for block in view_dict.get("blocks", [])
            if block.get("type") == "section"
        ]

    @staticmethod
    def _get_action_elements(view_dict: dict) -> list[dict]:
        elements: list[dict] = []
        for block in view_dict.get("blocks", []):
            if block.get("type") == "actions":
                elements.extend(block.get("elements", []))
        return elements

    @staticmethod
    def test_p1_p2_shows_auto_creation_message(mocker):
        incident = SimpleNamespace(id=123, needs_postmortem=True, status=IncidentStatus.OPEN)

        # No existing PMs
        mocker.patch(
            "firefighter.slack.views.modals.postmortem._safe_has_relation",
            return_value=False,
            create=True,
        )

        view = PostMortemModal().build_modal_fn(incident)
        view_dict = view.to_dict()

        texts = TestPostMortemModal._get_text_blocks(view_dict)
        assert any(
            "automatically created when the incident reaches MITIGATED" in t
            for t in texts
        )
        # No manual-create button for mandatory PMs
        action_ids = [
            el.get("action_id")
            for el in TestPostMortemModal._get_action_elements(view_dict)
        ]
        assert "incident_create_postmortem_now" not in action_ids

    @staticmethod
    def test_p1_p2_mitigated_shows_button(mocker):
        incident = SimpleNamespace(id=124, needs_postmortem=True, status=IncidentStatus.MITIGATED)

        mocker.patch(
            "firefighter.slack.views.modals.postmortem._safe_has_relation",
            return_value=False,
            create=True,
        )

        view = PostMortemModal().build_modal_fn(incident)
        view_dict = view.to_dict()

        texts = TestPostMortemModal._get_text_blocks(view_dict)
        assert any("not automatically created" in t for t in texts)

        action_ids = [
            el.get("action_id")
            for el in TestPostMortemModal._get_action_elements(view_dict)
        ]
        assert "incident_create_postmortem_now" in action_ids

    @staticmethod
    def test_p3_shows_optional_message_and_button(mocker):
        incident = SimpleNamespace(id=456, needs_postmortem=False)

        # No existing PMs
        mocker.patch(
            "firefighter.slack.views.modals.postmortem._safe_has_relation",
            return_value=False,
            create=True,
        )

        view = PostMortemModal().build_modal_fn(incident)
        view_dict = view.to_dict()

        texts = TestPostMortemModal._get_text_blocks(view_dict)
        assert any("P3 incident post-mortem is not mandatory" in t for t in texts)

        action_ids = [
            el.get("action_id")
            for el in TestPostMortemModal._get_action_elements(view_dict)
        ]
        assert "incident_create_postmortem_now" in action_ids



from __future__ import annotations

import pytest
from django.apps import apps

from firefighter.incidents.enums import IncidentStatus
from firefighter.incidents.factories import IncidentFactory, UserFactory
from firefighter.incidents.models import IncidentUpdate
from firefighter.jira_app.models import JiraPostMortem
from firefighter.slack.factories import IncidentChannelFactory
from firefighter.slack.messages.slack_messages import (
    SlackMessageDeployWarning,
    SlackMessageIncidentDeclaredAnnouncement,
    SlackMessageIncidentStatusUpdated,
)

try:
    from firefighter.confluence.models import PostMortem
except (ImportError, AttributeError):
    PostMortem = None

if not apps.is_installed("firefighter.confluence"):
    pytest.skip("Confluence app not installed; skipping slack message tests", allow_module_level=True)


@pytest.mark.django_db
class TestSlackMessageIncidentStatusUpdated:
    """Test SlackMessageIncidentStatusUpdated message generation."""

    def test_title_when_status_mitigated_and_changed(self) -> None:
        """Test that title mentions MITIGATED when incident reaches MITIGATED status.

        Covers line 445: elif incident.status == IncidentStatus.MITIGATED and status_changed:
        """
        # Create an incident in MITIGATED status
        incident = IncidentFactory.create(_status=IncidentStatus.MITIGATED)

        # Create an IncidentChannel for slack_channel_name
        IncidentChannelFactory.create(incident=incident)

        # Create an IncidentUpdate
        user = UserFactory.create()
        incident_update = IncidentUpdate.objects.create(
            incident=incident, status=IncidentStatus.MITIGATED, created_by=user
        )

        # Create the message with status_changed=True and in_channel=False
        message = SlackMessageIncidentStatusUpdated(
            incident=incident,
            incident_update=incident_update,
            in_channel=False,
            status_changed=True,
        )

        # Verify the title contains "Mitigated" (the status label)
        assert message.title_text is not None
        assert "Mitigated" in message.title_text
        assert ":large_green_circle:" in message.title_text
        assert incident.slack_channel_name in message.title_text

    def test_title_when_status_not_mitigated(self) -> None:
        """Test that title does not use MITIGATED-specific format when status is different."""
        # Create an incident in INVESTIGATING status
        incident = IncidentFactory.create(_status=IncidentStatus.INVESTIGATING)

        # Create an IncidentChannel
        IncidentChannelFactory.create(incident=incident)

        # Create an IncidentUpdate
        user = UserFactory.create()
        incident_update = IncidentUpdate.objects.create(
            incident=incident, status=IncidentStatus.INVESTIGATING, created_by=user
        )

        # Create the message with status_changed=True and in_channel=False
        message = SlackMessageIncidentStatusUpdated(
            incident=incident,
            incident_update=incident_update,
            in_channel=False,
            status_changed=True,
        )

        # Verify the title does NOT contain the MITIGATED-specific format
        assert message.title_text is not None
        assert ":large_green_circle:" not in message.title_text
        assert "has received an update" in message.title_text

    def test_no_update_button_when_incident_closed(self) -> None:
        """Test that Update Status button is NOT displayed when incident is CLOSED.

        This prevents showing an action button in an archived channel.
        """
        # Create an incident in CLOSED status
        incident = IncidentFactory.create(_status=IncidentStatus.CLOSED)

        # Create an IncidentChannel
        IncidentChannelFactory.create(incident=incident)

        # Create an IncidentUpdate with status change
        user = UserFactory.create()
        incident_update = IncidentUpdate.objects.create(
            incident=incident, status=IncidentStatus.CLOSED, created_by=user
        )

        # Create the message with in_channel=True (normal case for in-channel messages)
        message = SlackMessageIncidentStatusUpdated(
            incident=incident,
            incident_update=incident_update,
            in_channel=True,
            status_changed=True,
        )

        # Get the blocks
        blocks = message.get_blocks()

        # Find any SectionBlock with the "message_status_update" block_id
        status_update_block = None
        for block in blocks:
            if hasattr(block, "block_id") and block.block_id == "message_status_update":
                status_update_block = block
                break

        # If there is a status update block, verify it has NO accessory (button)
        if status_update_block:
            assert status_update_block.accessory is None, (
                "Update button should not be present when incident is CLOSED"
            )

    def test_update_button_shown_when_incident_not_closed(self) -> None:
        """Test that Update Status button IS displayed when incident is NOT CLOSED."""
        # Create an incident in INVESTIGATING status (not closed)
        incident = IncidentFactory.create(_status=IncidentStatus.INVESTIGATING)

        # Create an IncidentChannel
        IncidentChannelFactory.create(incident=incident)

        # Create an IncidentUpdate
        user = UserFactory.create()
        incident_update = IncidentUpdate.objects.create(
            incident=incident, status=IncidentStatus.INVESTIGATING, created_by=user
        )

        # Create the message with in_channel=True
        message = SlackMessageIncidentStatusUpdated(
            incident=incident,
            incident_update=incident_update,
            in_channel=True,
            status_changed=True,
        )

        # Get the blocks
        blocks = message.get_blocks()

        # Find the SectionBlock with "message_status_update" block_id
        status_update_block = None
        for block in blocks:
            if hasattr(block, "block_id") and block.block_id == "message_status_update":
                status_update_block = block
                break

        # Verify the block has an accessory (Update button)
        assert status_update_block is not None, "Should have a status update block"
        assert status_update_block.accessory is not None, (
            "Update button should be present when incident is not CLOSED"
        )
        assert status_update_block.accessory.text.text == "Update"


@pytest.mark.django_db
class TestSlackMessageDeployWarning:
    """Test SlackMessageDeployWarning message generation."""

    def test_header_when_incident_mitigated(self) -> None:
        """Test that header includes '(Mitigated)' when incident is MITIGATED.

        Covers line 694: text=f":warning: Deploy warning {'(Mitigated) ' if self.incident.status == IncidentStatus.MITIGATED else ''}:warning:"
        """
        # Create an incident in MITIGATED status
        incident = IncidentFactory.create(_status=IncidentStatus.MITIGATED)

        # Create an IncidentChannel for conversation.name
        IncidentChannelFactory.create(incident=incident)

        # Create the deploy warning message
        message = SlackMessageDeployWarning(incident=incident)

        # Get the blocks
        blocks = message.get_blocks()

        # The first block should be a HeaderBlock with "(Mitigated)" in the text
        header_block = blocks[0]
        header_text = header_block.text.text

        assert "(Mitigated)" in header_text
        assert ":warning:" in header_text

    def test_header_when_incident_not_mitigated(self) -> None:
        """Test that header does NOT include '(Mitigated)' when incident is not MITIGATED."""
        # Create an incident in INVESTIGATING status
        incident = IncidentFactory.create(_status=IncidentStatus.INVESTIGATING)

        # Create an IncidentChannel
        IncidentChannelFactory.create(incident=incident)

        # Create the deploy warning message
        message = SlackMessageDeployWarning(incident=incident)

        # Get the blocks
        blocks = message.get_blocks()

        # The first block should be a HeaderBlock WITHOUT "(Mitigated)"
        header_block = blocks[0]
        header_text = header_block.text.text

        assert "(Mitigated)" not in header_text
        assert ":warning:" in header_text

    def test_additional_blocks_when_incident_mitigated_or_above(self) -> None:
        """Test that additional blocks are added when incident status >= MITIGATED.

        Covers line 705: if self.incident.status >= IncidentStatus.MITIGATED:
        """
        # Create an incident in MITIGATED status
        incident = IncidentFactory.create(_status=IncidentStatus.MITIGATED)

        # Create an IncidentChannel
        IncidentChannelFactory.create(incident=incident)

        # Create the deploy warning message
        message = SlackMessageDeployWarning(incident=incident)

        # Get the blocks
        blocks = message.get_blocks()

        # When status >= MITIGATED, there should be MORE than 2 blocks
        # (HeaderBlock + SectionBlock + additional blocks from line 706)
        assert len(blocks) > 2

    def test_no_additional_blocks_when_incident_below_mitigated(self) -> None:
        """Test that NO additional blocks are added when incident status < MITIGATED."""
        # Create an incident in INVESTIGATING status (below MITIGATED)
        incident = IncidentFactory.create(_status=IncidentStatus.INVESTIGATING)

        # Create an IncidentChannel
        IncidentChannelFactory.create(incident=incident)

        # Create the deploy warning message
        message = SlackMessageDeployWarning(incident=incident)

        # Get the blocks
        blocks = message.get_blocks()

        # When status < MITIGATED, there should be exactly 2 blocks
        # (HeaderBlock + SectionBlock only, no additional blocks)
        assert len(blocks) == 2


@pytest.mark.django_db
class TestSlackMessageIncidentDeclaredAnnouncement:
    """Test SlackMessageIncidentDeclaredAnnouncement message with custom fields."""

    def test_custom_fields_displayed_when_present(self) -> None:
        """Test that custom fields are displayed in the incident announcement."""
        # Create an incident with custom fields
        incident = IncidentFactory.create(
            custom_fields={
                "zendesk_ticket_id": "12345",
                "seller_contract_id": "SELLER-67890",
                "zoho_desk_ticket_id": "ZD-11111",
                "is_key_account": True,
                "is_seller_in_golden_list": True,
            }
        )

        # Create the message
        message = SlackMessageIncidentDeclaredAnnouncement(incident=incident)

        # Get the blocks
        blocks = message.get_blocks()

        # Find the SectionBlock with fields (should contain custom fields)
        # This is typically the 4th block (index 3)
        fields_block = None
        for block in blocks:
            if hasattr(block, "fields") and block.fields:
                fields_block = block
                break

        assert fields_block is not None, "Should have a block with fields"

        # Convert fields to strings for easier assertion (access .text attribute)
        fields_text = " ".join(
            field.text if hasattr(field, "text") else str(field)
            for field in fields_block.fields
        )

        # Verify custom fields are present
        assert "Zendesk Ticket" in fields_text
        assert "12345" in fields_text
        assert "Seller Contract" in fields_text
        assert "SELLER-67890" in fields_text
        assert "Zoho Desk Ticket" in fields_text
        assert "ZD-11111" in fields_text
        assert "Key Account" in fields_text
        assert "Golden List Seller" in fields_text

    def test_custom_fields_not_displayed_when_absent(self) -> None:
        """Test that custom fields are NOT displayed when not present."""
        # Create an incident WITHOUT custom fields
        incident = IncidentFactory.create(custom_fields={})

        # Create the message
        message = SlackMessageIncidentDeclaredAnnouncement(incident=incident)

        # Get the blocks
        blocks = message.get_blocks()

        # Find the SectionBlock with fields
        fields_block = None
        for block in blocks:
            if hasattr(block, "fields") and block.fields:
                fields_block = block
                break

        assert fields_block is not None, "Should have a block with fields"

        # Convert fields to strings (access .text attribute)
        fields_text = " ".join(
            field.text if hasattr(field, "text") else str(field)
            for field in fields_block.fields
        )

        # Verify custom fields are NOT present
        assert "Zendesk Ticket" not in fields_text
        assert "Seller Contract" not in fields_text
        assert "Zoho Desk Ticket" not in fields_text
        assert "Key Account" not in fields_text
        assert "Golden List Seller" not in fields_text

    def test_partial_custom_fields_displayed(self) -> None:
        """Test that only filled custom fields are displayed."""
        # Create an incident with only some custom fields
        incident = IncidentFactory.create(
            custom_fields={
                "zendesk_ticket_id": "12345",
                # seller_contract_id not set
                # zoho_desk_ticket_id not set
                "is_key_account": False,  # False should not display
                # is_seller_in_golden_list not set
            }
        )

        # Create the message
        message = SlackMessageIncidentDeclaredAnnouncement(incident=incident)

        # Get the blocks
        blocks = message.get_blocks()

        # Find the SectionBlock with fields
        fields_block = None
        for block in blocks:
            if hasattr(block, "fields") and block.fields:
                fields_block = block
                break

        assert fields_block is not None, "Should have a block with fields"

        # Convert fields to strings (access .text attribute)
        fields_text = " ".join(
            field.text if hasattr(field, "text") else str(field)
            for field in fields_block.fields
        )

        # Verify only zendesk_ticket_id is present
        assert "Zendesk Ticket" in fields_text
        assert "12345" in fields_text

        # Verify others are NOT present
        assert "Seller Contract" not in fields_text
        assert "Zoho Desk Ticket" not in fields_text
        assert "Key Account" not in fields_text
        assert "Golden List Seller" not in fields_text

    def test_jira_postmortem_link_displayed_when_present(self) -> None:
        """Test that Jira post-mortem link is displayed when available."""
        # Create an incident
        incident = IncidentFactory.create()
        user = UserFactory.create()

        # Create a Jira post-mortem for the incident
        jira_pm = JiraPostMortem.objects.create(
            incident=incident,
            jira_issue_key="PM-123",
            jira_issue_id="12345",
            created_by=user,
        )

        # Create the message
        message = SlackMessageIncidentDeclaredAnnouncement(incident=incident)

        # Get the blocks
        blocks = message.get_blocks()

        # Collect all fields from all SectionBlocks
        all_fields = []
        for block in blocks:
            if hasattr(block, "fields") and block.fields:
                all_fields.extend(block.fields)

        assert len(all_fields) > 0, "Should have blocks with fields"

        # Convert fields to strings (access .text attribute)
        fields_text = " ".join(
            field.text if hasattr(field, "text") else str(field)
            for field in all_fields
        )

        # Verify Jira post-mortem link is present (now in separate PM block)
        assert "Jira Post-mortem" in fields_text
        assert "PM-123" in fields_text
        assert jira_pm.issue_url in fields_text

    def test_confluence_postmortem_link_displayed_when_present(self) -> None:
        """Test that Confluence post-mortem link is displayed when available."""
        if PostMortem is None:
            pytest.skip("Confluence app not installed")

        # Create an incident
        incident = IncidentFactory.create()

        # Create a Confluence post-mortem for the incident
        confluence_pm = PostMortem.objects.create(
            incident=incident,
            page_id="123456",
        )

        # Create the message
        message = SlackMessageIncidentDeclaredAnnouncement(incident=incident)

        # Get the blocks
        blocks = message.get_blocks()

        # Find the SectionBlock with fields
        fields_block = None
        for block in blocks:
            if hasattr(block, "fields") and block.fields:
                fields_block = block
                break

        assert fields_block is not None, "Should have a block with fields"

        # Convert fields to strings (access .text attribute)
        fields_text = " ".join(
            field.text if hasattr(field, "text") else str(field)
            for field in fields_block.fields
        )

        # Verify Confluence post-mortem link is present
        assert "Confluence Post-mortem" in fields_text
        assert confluence_pm.page_url in fields_text



from __future__ import annotations

import uuid

from django.db import models
from django.db.models.manager import Manager


class Environment(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    value = models.CharField(max_length=16, unique=True)
    name = models.CharField(max_length=128, unique=True)
    description = models.CharField(max_length=128, unique=True)
    order = models.IntegerField(default=0)
    default = models.BooleanField(default=False)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    objects = Manager["Environment"]()

    def __str__(self) -> str:
        return self.value

    @classmethod
    def get_default(cls) -> Environment:
        return Environment.objects.get(default=True)



from __future__ import annotations

from rest_framework import serializers

from firefighter.confluence.models import Runbook


class RunbookSerializer(serializers.ModelSerializer[Runbook]):
    class Meta:
        model = Runbook
        fields = "__all__"



"""Signal handlers for incident key events updates to sync with Jira post-mortem."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from django.conf import settings
from django.dispatch import receiver
from django.template.loader import render_to_string

from firefighter.incidents.models.incident import Incident as IncidentModel
from firefighter.incidents.signals import incident_key_events_updated
from firefighter.jira_app.client import JiraClient
from firefighter.jira_app.service_postmortem import JiraPostMortemService

if TYPE_CHECKING:
    from firefighter.incidents.models.incident import Incident

logger = logging.getLogger(__name__)


@receiver(signal=incident_key_events_updated)
def sync_key_events_to_jira_postmortem(
    sender: Any, incident: Incident, **kwargs: dict[str, Any]
) -> None:
    """Update Jira post-mortem timeline when key events are updated.

    This handler is triggered when incident key events are updated via the web UI
    or Slack, and syncs the timeline to the associated Jira post-mortem ticket.

    Args:
        sender: The sender of the signal
        incident: The incident whose key events were updated
        **kwargs: Additional keyword arguments
    """
    # Check if Jira post-mortem is enabled
    if not getattr(settings, "ENABLE_JIRA_POSTMORTEM", False):
        logger.debug("Jira post-mortem disabled, skipping timeline sync")
        return

    # Check if incident has a Jira post-mortem
    if not hasattr(incident, "jira_postmortem_for") or not incident.jira_postmortem_for:
        logger.debug(f"Incident #{incident.id} has no Jira post-mortem, skipping timeline sync")
        return

    jira_postmortem = incident.jira_postmortem_for
    logger.info(
        f"Syncing key events timeline to Jira post-mortem {jira_postmortem.jira_issue_key} "
        f"for incident #{incident.id}"
    )

    try:
        # Prefetch incident updates for timeline generation
        incident_refreshed = (
            IncidentModel.objects.select_related("priority", "environment")
            .prefetch_related("incidentupdate_set")
            .get(pk=incident.pk)
        )

        # Generate updated timeline from template
        timeline_content = render_to_string(
            "jira/postmortem/timeline.txt",
            {"incident": incident_refreshed},
        )

        # Get the field ID for timeline from service
        service = JiraPostMortemService()
        timeline_field_id = service.field_ids.get("timeline")

        if not timeline_field_id:
            logger.error("Timeline field ID not found in Jira post-mortem service configuration")
            return

        # Update the Jira ticket
        client = JiraClient()
        issue = client.jira.issue(jira_postmortem.jira_issue_key)
        issue.update(fields={timeline_field_id: timeline_content})

        logger.info(
            f"Successfully updated timeline in Jira post-mortem {jira_postmortem.jira_issue_key}"
        )

    except Exception:
        logger.exception(
            f"Failed to update timeline in Jira post-mortem {jira_postmortem.jira_issue_key} "
            f"for incident #{incident.id}"
        )



from __future__ import annotations

from typing import Any, TypeVar

from django.db.models import Model
from drf_spectacular.utils import OpenApiExample, OpenApiParameter, extend_schema
from rest_framework import mixins, viewsets

T_co = TypeVar("T_co", bound=Model, covariant=True)


@extend_schema(
    parameters=[
        OpenApiParameter(
            name="fields",
            type=str,
            location=OpenApiParameter.QUERY,
            description="Comma-separated list of fields to include in a CSV or TSV render. Nested objects are de-nested and accessible with a dot `.`",
            default="__all__",
            examples=[
                OpenApiExample(
                    name="Comma separated list of fields",
                    value="id,name,incident_category.name,incident_category.group.name",
                    request_only=True,
                ),
                OpenApiExample(
                    name="All fields (default)",
                    value="__all__",
                    request_only=True,
                ),
            ],
        )
    ],
)
class AdvancedGenericViewSet[T_co: Model](viewsets.GenericViewSet[T_co]):
    """Generic viewset with support for fields and labels parameters, to customize fields of CSV and TSV renderer."""

    fields: list[str] = []
    """List of attributes/properties to be rendered in CSV/TSV format.
    For nested objects, a dot is used a separator.

     E.g. `fields = ["id", "name", "component.name", "component.group.name"]`
    """

    labels: dict[str, str] = {}
    """
    Dict mapping attributes/properties to a more friendly string to be shown in the header row for CSV/TSV renders.

    E.g. `labels = {"slack_channel_name": "Slack Channel"}`
    """

    def get_renderer_context(self) -> dict[str, Any]:
        context = super().get_renderer_context()

        if self.request.GET.get("fields") == "__all__":
            context["header"] = None
        else:
            context["header"] = (
                self.request.GET["fields"].split(",")
                if "fields" in self.request.GET
                else self.fields
            )
        if self.request.GET.get("labels") == "__original__":
            context["labels"] = None
        else:
            context["labels"] = (
                "__hidden__"
                if self.request.GET.get("labels") == "__hidden__"
                else self.labels
            )
        return context


class ReadOnlyModelViewSet[T_co: Model](
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    AdvancedGenericViewSet[T_co],
):
    pass



from datetime import timedelta

from django.db import migrations


def update_priority_settings(apps, schema_editor):
    Priority = apps.get_model("incidents", "Priority")
    priorities_to_update = {
        "P1": (timedelta(hours=2), True, True),
        "P2": (timedelta(hours=4), True, True),
        "P3": (timedelta(days=1), True, True),
        "P4": (timedelta(days=5), True, True),
        "P5": (timedelta(days=10), True, True),
    }

    for name, (sla, enabled_create, enabled_update) in priorities_to_update.items():
        Priority.objects.filter(name=name).update(
            sla=sla,
            enabled_create=enabled_create,
            enabled_update=enabled_update,
        )


class Migration(migrations.Migration):

    dependencies = [
        ("incidents", "0008_impact_level"),
    ]

    operations = [
        migrations.RunPython(update_priority_settings),
    ]



"""Tests for ClosureReasonModal - Slack modal for closing incidents with a reason."""

from __future__ import annotations

from unittest.mock import MagicMock, patch

import pytest
from slack_sdk.errors import SlackApiError

from firefighter.incidents.enums import ClosureReason, IncidentStatus
from firefighter.incidents.factories import IncidentFactory, UserFactory
from firefighter.incidents.models import Environment, Priority
from firefighter.slack.views.modals.closure_reason import ClosureReasonModal


@pytest.mark.django_db
class TestClosureReasonModalMessageTabDisabled:
    """Test ClosureReasonModal handles messages_tab_disabled gracefully."""

    def test_closure_reason_handles_messages_tab_disabled(
        self, caplog: pytest.LogCaptureFixture, mocker
    ) -> None:
        """Test that messages_tab_disabled error is handled gracefully with warning log."""
        # Create test data
        user = UserFactory.build()
        user.save()
        incident = IncidentFactory.build(
            _status=IncidentStatus.INVESTIGATING, created_by=user
        )
        incident.save()

        # Mock can_be_closed to return True so the closure can proceed
        mocker.patch.object(
            type(incident),
            "can_be_closed",
            new_callable=mocker.PropertyMock,
            return_value=(True, []),
        )

        # Create modal and mock
        modal = ClosureReasonModal()
        ack = MagicMock()

        # Create valid form submission
        body = {
            "view": {
                "state": {
                    "values": {
                        "closure_reason": {
                            "select_closure_reason": {
                                "selected_option": {"value": ClosureReason.CANCELLED}
                            }
                        },
                        "closure_reference": {"input_closure_reference": {"value": ""}},
                        "closure_message": {
                            "input_closure_message": {"value": "Test closure message"}
                        },
                    }
                },
                "private_metadata": str(incident.id),
            },
            "user": {"id": "U123456"},
        }

        # Mock respond to raise messages_tab_disabled error
        slack_error_response = MagicMock()
        slack_error_response.get.return_value = "messages_tab_disabled"

        with patch(
            "firefighter.slack.views.modals.closure_reason.respond"
        ) as mock_respond:
            mock_respond.side_effect = SlackApiError(
                message="The request to the Slack API failed.",
                response=slack_error_response,
            )

            # Execute
            result = modal.handle_modal_fn(
                ack=ack, body=body, incident=incident, user=user
            )

            # Assertions
            assert result is True  # Incident should still be closed successfully

            # Verify incident was closed
            incident.refresh_from_db()
            assert incident.status == IncidentStatus.CLOSED
            assert incident.closure_reason == ClosureReason.CANCELLED

            # Verify warning was logged
            assert any(
                "Cannot send DM to user" in record.message
                and record.levelname == "WARNING"
                for record in caplog.records
            )

    def test_closure_reason_reraises_other_slack_errors(self, mocker) -> None:
        """Test that other Slack API errors are re-raised."""
        # Create test data
        user = UserFactory.build()
        user.save()
        incident = IncidentFactory.build(
            _status=IncidentStatus.INVESTIGATING, created_by=user
        )
        incident.save()

        # Mock can_be_closed to return True so the closure can proceed
        mocker.patch.object(
            type(incident),
            "can_be_closed",
            new_callable=mocker.PropertyMock,
            return_value=(True, []),
        )

        # Create modal and mock
        modal = ClosureReasonModal()
        ack = MagicMock()

        # Create valid form submission
        body = {
            "view": {
                "state": {
                    "values": {
                        "closure_reason": {
                            "select_closure_reason": {
                                "selected_option": {"value": ClosureReason.CANCELLED}
                            }
                        },
                        "closure_reference": {"input_closure_reference": {"value": ""}},
                        "closure_message": {
                            "input_closure_message": {"value": "Test closure message"}
                        },
                    }
                },
                "private_metadata": str(incident.id),
            },
            "user": {"id": "U123456"},
        }

        # Mock respond to raise different Slack error
        slack_error_response = MagicMock()
        slack_error_response.get.return_value = "channel_not_found"

        with patch(
            "firefighter.slack.views.modals.closure_reason.respond"
        ) as mock_respond:
            mock_respond.side_effect = SlackApiError(
                message="The request to the Slack API failed.",
                response=slack_error_response,
            )

            # Execute and expect exception
            with pytest.raises(SlackApiError):
                modal.handle_modal_fn(ack=ack, body=body, incident=incident, user=user)


@pytest.mark.django_db
class TestClosureReasonModalEarlyClosureBypass:
    """Test early-closure path respects submitted closure reason."""

    def test_allows_early_closure_with_submitted_reason(self, settings) -> None:
        """Ensure can_be_closed passes when a closure reason is provided for early closure."""
        settings.ENABLE_JIRA_POSTMORTEM = True

        # Ensure required priority/environment for needs_postmortem + PRD
        priority = Priority.objects.create(
            name="P1-test",
            value=9991,
            description="P1 test",
            order=9991,
            needs_postmortem=True,
        )
        env, _ = Environment.objects.get_or_create(
            value="PRD",
            defaults={
                "name": "Production",
                "description": "Production",
                "order": 9991,
            },
        )

        user = UserFactory.create()
        incident = IncidentFactory.create(
            _status=IncidentStatus.INVESTIGATING,
            created_by=user,
            priority=priority,
            environment=env,
        )

        modal = ClosureReasonModal()
        ack = MagicMock()

        body = {
            "view": {
                "state": {
                    "values": {
                        "closure_reason": {
                            "select_closure_reason": {
                                "selected_option": {"value": ClosureReason.CANCELLED}
                            }
                        },
                        "closure_reference": {
                            "input_closure_reference": {"value": "INC-42"}
                        },
                        "closure_message": {
                            "input_closure_message": {
                                "value": "Closing early with reason"
                            }
                        },
                    }
                },
                "private_metadata": str(incident.id),
            },
            "user": {"id": "U123456"},
        }

        with patch(
            "firefighter.slack.views.modals.closure_reason.respond"
        ) as mock_respond:
            mock_respond.return_value = None

            result = modal.handle_modal_fn(
                ack=ack,
                body=body,
                incident=incident,
                user=user,
            )

        # Early closure should succeed and close the incident
        assert result is True
        incident.refresh_from_db()
        assert incident.status == IncidentStatus.CLOSED
        assert incident.closure_reason == ClosureReason.CANCELLED
        assert incident.closure_reference == "INC-42"

        # Ack should clear modal stack
        ack.assert_called_once_with(response_action="clear")



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from django import forms

from firefighter.incidents.enums import IncidentStatus
from firefighter.incidents.forms.utils import EnumChoiceField, GroupedModelChoiceField
from firefighter.incidents.models import IncidentCategory, Priority

if TYPE_CHECKING:
    from firefighter.incidents.models import Incident

logger = logging.getLogger(__name__)


class UpdateStatusForm(forms.Form):
    message = forms.CharField(
        label="Update message",
        widget=forms.Textarea,
        min_length=10,
        max_length=1200,
        required=False,
    )
    status = EnumChoiceField(
        enum_class=IncidentStatus,
        label="Status",
        choices=IncidentStatus.choices_lte(IncidentStatus.CLOSED),
    )
    priority = forms.ModelChoiceField(
        label="Priority",
        queryset=Priority.objects.filter(enabled_update=True),
    )
    incident_category = GroupedModelChoiceField(
        choices_groupby="group",
        label="Incident category",
        queryset=IncidentCategory.objects.all()
        .select_related("group")
        .order_by(
            "group__order",
            "name",
        ),
    )

    def __init__(
        self, *args: Any, incident: Incident | None = None, **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)

        # Store incident for later use in clean()
        self.incident = incident

        # Dynamically adjust status choices based on incident requirements
        if incident:
            self._set_status_choices(incident)

    def _set_status_choices(self, incident: Incident) -> None:
        """Set the status field choices based on the incident's current state and requirements."""
        current_status = incident.status
        status_field = self.fields["status"]

        # Check if incident requires post-mortem (P1/P2 in PRD)
        requires_postmortem = bool(
            incident.priority
            and incident.environment
            and incident.priority.needs_postmortem
            and incident.environment.value == "PRD"
        )
        allowed_statuses = self._get_allowed_statuses(
            current_status, requires_postmortem=requires_postmortem
        )

        # If we got a list of enum values, convert to choices and include current status
        if allowed_statuses:
            if current_status not in allowed_statuses:
                allowed_statuses.insert(0, current_status)
            # Convert values to strings to match what Slack sends in form submissions
            choices = [(str(s.value), s.label) for s in allowed_statuses]
            status_field.choices = choices  # type: ignore[attr-defined]

    def _get_allowed_statuses(
        self, current_status: IncidentStatus, *, requires_postmortem: bool
    ) -> list[IncidentStatus] | None:
        """Get allowed status transitions based on current status and postmortem requirement.

        Returns None if choices should be set directly (for default fallback cases).
        """
        # For incidents requiring post-mortem (P1/P2 in PRD)
        if requires_postmortem:
            return self._get_postmortem_allowed_statuses(current_status)

        # For P3+ incidents (no post-mortem needed)
        return self._get_no_postmortem_allowed_statuses(current_status)

    def _get_postmortem_allowed_statuses(
        self, current_status: IncidentStatus
    ) -> list[IncidentStatus] | None:
        """Get allowed statuses for incidents requiring postmortem (P1/P2)."""
        if current_status == IncidentStatus.OPEN:
            return [IncidentStatus.INVESTIGATING, IncidentStatus.CLOSED]
        if current_status == IncidentStatus.INVESTIGATING:
            return [IncidentStatus.MITIGATING, IncidentStatus.CLOSED]
        if current_status == IncidentStatus.MITIGATING:
            return [IncidentStatus.MITIGATED]
        if current_status == IncidentStatus.MITIGATED:
            # P1/P2 can: go to POST_MORTEM (required) OR reopen to INVESTIGATING/MITIGATING (with reason)
            return [
                IncidentStatus.POST_MORTEM,      # Required next step
                IncidentStatus.INVESTIGATING,   # Reopen option (with reason)
                IncidentStatus.MITIGATING,      # Reopen option (with reason)
            ]
        if current_status == IncidentStatus.POST_MORTEM:
            return [IncidentStatus.CLOSED]

        # For any other status, return None to use the default choices
        return None

    def _get_no_postmortem_allowed_statuses(
        self, current_status: IncidentStatus
    ) -> list[IncidentStatus] | None:
        """Get allowed statuses for incidents not requiring postmortem (P3/P4/P5)."""
        if current_status == IncidentStatus.OPEN:
            return [IncidentStatus.INVESTIGATING, IncidentStatus.CLOSED]
        if current_status == IncidentStatus.INVESTIGATING:
            return [IncidentStatus.MITIGATING, IncidentStatus.CLOSED]
        if current_status == IncidentStatus.MITIGATING:
            return [IncidentStatus.MITIGATED]
        if current_status == IncidentStatus.MITIGATED:
            # P3/P4/P5 can: go to CLOSED (normal next step) OR reopen to INVESTIGATING/MITIGATING (with reason)
            return [
                IncidentStatus.CLOSED,          # Normal next step for P3+
                IncidentStatus.INVESTIGATING,   # Reopen option (with reason)
                IncidentStatus.MITIGATING,      # Reopen option (with reason)
            ]

        # For any other status, return None to use the default choices
        return None

    def _set_default_choices(
        self, status_field: Any, current_status: IncidentStatus, default_choices: Any
    ) -> None:
        """Set status field choices to default, ensuring current status is included."""
        # Convert default_choices to string keys to match Slack form submissions
        status_field.choices = [
            (str(choice[0]), choice[1]) for choice in default_choices
        ]
        existing_values = {choice[0] for choice in status_field.choices}
        if str(current_status.value) not in existing_values:
            # Insert current status at the beginning
            status_field.choices = [
                (str(current_status.value), current_status.label),
                *status_field.choices,
            ]

    @staticmethod
    def requires_closure_reason(
        incident: Incident, target_status: IncidentStatus
    ) -> bool:
        """Check if closing this incident to the target status requires a closure reason.

        Based on the workflow diagram:
        - P1/P2 and P3/P4/P5: require reason when closing from Opened or Investigating
        """
        if target_status != IncidentStatus.CLOSED:
            return False

        current_status = incident.status

        # Require reason if closing from Opened or Investigating (for any priority)
        return current_status.value in {
            IncidentStatus.OPEN,
            IncidentStatus.INVESTIGATING,
        }

    def clean_message(self) -> str:
        """Validate message field, ensuring reopening from MITIGATED has sufficient justification."""
        message = self.cleaned_data.get("message", "").strip()

        # Get the incident and status from form data/initialization
        incident = getattr(self, "incident", None)
        status_value = self.data.get("status")  # Use raw form data as cleaned_data may not be ready yet

        if incident and status_value:
            current_status = incident.status

            # Convert status value to enum if it's a string/number
            try:
                if isinstance(status_value, str):
                    status = IncidentStatus(int(status_value))
                else:
                    status = IncidentStatus(status_value)
            except (ValueError, TypeError):
                # Invalid status value, let other validation handle it
                return message

            # If reopening from MITIGATED to earlier phases, require substantial justification
            if (current_status == IncidentStatus.MITIGATED and
                status in {IncidentStatus.INVESTIGATING, IncidentStatus.MITIGATING} and
                len(message) < 10):
                raise forms.ValidationError(
                    "A detailed justification (minimum 10 characters) is required when reopening "
                    "an incident from MITIGATED status."
                )

        return message



from __future__ import annotations

import logging
from datetime import timedelta
from typing import TYPE_CHECKING, Any, TypeVar

from django.core.exceptions import ObjectDoesNotExist
from django.core.validators import EmailValidator
from django.db.models import Model, QuerySet
from drf_spectacular.extensions import (
    OpenApiSerializerExtension,
)
from rest_framework import serializers
from rest_framework.fields import empty
from taggit.serializers import TaggitSerializer, TagListSerializerField

from firefighter.firefighter.utils import get_in
from firefighter.incidents.models.environment import Environment
from firefighter.incidents.models.group import Group
from firefighter.incidents.models.incident import Incident
from firefighter.incidents.models.incident_category import IncidentCategory
from firefighter.incidents.models.incident_cost import IncidentCost
from firefighter.incidents.models.incident_cost_type import IncidentCostType
from firefighter.incidents.models.incident_membership import IncidentRole
from firefighter.incidents.models.metric_type import IncidentMetric, MetricType
from firefighter.incidents.models.priority import Priority
from firefighter.incidents.models.user import User

if TYPE_CHECKING:
    from collections.abc import Callable, MutableMapping, Sequence

    from drf_spectacular.openapi import AutoSchema
    from drf_spectacular.utils import Direction


logger = logging.getLogger(__name__)

T = TypeVar("T", bound=Model)


class GroupedModelSerializer[T: Model](serializers.RelatedField[T, Any, Any]):
    """Generic implementation for a model with a One2Many, where instead of a list we want the items grouped by a field, like costs or metrics."""

    def __init__(
        self,
        child_serializer: type[serializers.ModelSerializer[T]],
        key_getter: Callable[[T], str] = lambda x: str(x.pk),
        **kwargs: Any,
    ):
        self.child_serializer = child_serializer
        self.key_getter = key_getter
        if not kwargs.get("read_only", True):
            raise ValueError("GroupedModelSerializer can only be read_only")
        if "read_only" not in kwargs:
            kwargs["read_only"] = True
        kwargs["many"] = True
        super().__init__(**kwargs)

    def to_representation(self, value: Sequence[T]) -> dict[str, Any]:  # type: ignore[override]
        return {
            self.key_getter(child): self.child_serializer(child).data for child in value
        }


class GroupedModelSerializerOpenAPI(OpenApiSerializerExtension):  # type: ignore[no-untyped-call]
    target_class = "api.serializers.GroupedModelSerializer"
    target: GroupedModelSerializer[Any]

    def map_serializer(
        self, auto_schema: AutoSchema, direction: Direction
    ) -> dict[str, Any]:
        child_schema = auto_schema.resolve_serializer(
            self.target.child_serializer, direction
        )
        child_schema_ref = child_schema.ref or child_schema
        return {"type": "object", "additionalProperties": child_schema_ref}

    def get_name(
        self,
        auto_schema: AutoSchema,  # noqa: ARG002
        direction: Direction,  # noqa: ARG002
    ) -> str:
        return f"GroupedModelSerializer_{self.target.child_serializer.__name__}"


class CreatableSlugRelatedField[T: Model](serializers.SlugRelatedField[T]):
    """Like [SlugRelatedField](https://www.django-rest-framework.org/api-guide/relations/#slugrelatedfield), but allows to create an object that is not present in DB."""

    def to_internal_value(self, data: str | None) -> T:
        if self.slug_field is None:
            raise ValueError(
                "SlugRelatedField requires the `slug_field` attribute to be set."
            )

        try:
            data_get_or_create: MutableMapping[str, Any] = {self.slug_field: data}
        except ObjectDoesNotExist:
            logger.warning("Object does not exist. Creating...")
            # XXX Improve behaviour when user does not exist
            self.get_queryset().create(**{self.slug_field: data})
        except (TypeError, ValueError):
            return self.fail("invalid")
        else:
            queryset: QuerySet[T] = self.get_queryset()
            return queryset.get(**data_get_or_create)
        return self.fail("invalid")

    def run_validation(self, data: Any = empty) -> T | Any | None:
        """Override the default validation to perform the validation before trying to create the object."""
        (is_empty_value, data) = self.validate_empty_values(data)
        if is_empty_value:
            return data
        self.run_validators(data)
        return self.to_internal_value(data)


class DurationFieldSeconds(serializers.DurationField):
    def to_representation(self, value: timedelta) -> float:  # type: ignore[override]
        return value.total_seconds()

    def to_internal_value(self, value: float) -> timedelta:  # type: ignore[override]
        return timedelta(seconds=value)


class EnvironmentSerializer(serializers.ModelSerializer[Environment]):
    class Meta:
        model = Environment
        exclude = ["created_at", "updated_at"]


class IncidentRoleSerializer(serializers.ModelSerializer[IncidentRole]):
    class Meta:
        model = IncidentRole
        fields = "__all__"


class IncidentRoleInlineSerializer(IncidentRoleSerializer):
    name = serializers.StringRelatedField[User](read_only=True, source="user.name")
    email = serializers.StringRelatedField[User](read_only=True, source="user.email")
    id = serializers.PrimaryKeyRelatedField[User](read_only=True, source="user.id")

    class Meta:
        model = IncidentRole
        fields = ["id", "name", "email"]


class PrioritySerializer(serializers.ModelSerializer[Priority]):
    class Meta:
        depth = 1
        model = Priority
        fields = "__all__"


class GroupSerializer(serializers.ModelSerializer[Group]):
    class Meta:
        model = Group
        fields = "__all__"


class IncidentCategorySerializer(serializers.ModelSerializer[IncidentCategory]):
    group = GroupSerializer()

    class Meta:
        model = IncidentCategory
        fields = "__all__"


class UserSerializer(serializers.ModelSerializer[User]):
    class Meta:
        model = User
        fields = ["id", "name", "email"]


class IncidentCostTypeSerializer(serializers.ModelSerializer[IncidentCostType]):
    class Meta:
        model = IncidentCostType
        fields = "__all__"


class IncidentCostSerializer(serializers.ModelSerializer[IncidentCost]):
    cost_type = IncidentCostTypeSerializer()

    class Meta:
        model = IncidentCost
        fields = "__all__"


class IncidentCostSerializerInline(IncidentCostSerializer):
    class Meta:
        model = IncidentCost
        fields = ["amount", "details"]


class IncidentMetricSerializer(serializers.ModelSerializer[IncidentMetric]):
    duration = serializers.DurationField(read_only=True)
    duration_seconds = DurationFieldSeconds(read_only=True, source="duration")
    metric_type = serializers.StringRelatedField[MetricType](read_only=True)

    class Meta:
        model = IncidentMetric
        exclude = ["id", "incident"]


class IncidentSerializer(TaggitSerializer, serializers.ModelSerializer[Incident]):
    environment = EnvironmentSerializer(read_only=True)
    environment_id = serializers.PrimaryKeyRelatedField(
        source="environment", queryset=Environment.objects.all(), write_only=True
    )
    priority = PrioritySerializer(read_only=True)
    priority_id = serializers.PrimaryKeyRelatedField(
        source="priority", queryset=Priority.objects.all(), write_only=True
    )
    incident_category = IncidentCategorySerializer(read_only=True)
    incident_category_id = serializers.PrimaryKeyRelatedField(
        source="incident_category", queryset=IncidentCategory.objects.all(), write_only=True
    )

    status = serializers.SerializerMethodField()
    created_by = UserSerializer(read_only=True)
    slack_channel_name = serializers.SerializerMethodField()
    postmortem_url = serializers.SerializerMethodField()
    jira_ticket_key = serializers.SerializerMethodField()
    jira_ticket_url = serializers.SerializerMethodField()

    created_by_email = CreatableSlugRelatedField[User](
        source="created_by",
        write_only=True,
        slug_field="email",
        queryset=User.objects.all(),
        validators=[EmailValidator()],  # type: ignore[list-item]
    )

    tags = TagListSerializerField(read_only=True)
    metrics = GroupedModelSerializer(
        source="metrics_prefetched",
        child_serializer=IncidentMetricSerializer,
        key_getter=lambda x: x.metric_type.type,
    )
    costs = GroupedModelSerializer[IncidentCost](
        source="costs_prefetched",
        child_serializer=IncidentCostSerializerInline,
        key_getter=lambda x: x.cost_type.name,
    )
    roles = GroupedModelSerializer[IncidentRole](
        source="roles_prefetched",
        child_serializer=IncidentRoleInlineSerializer,
        key_getter=lambda x: str(x.role_type.slug),
    )

    @staticmethod
    def get_status(obj: Incident) -> str:
        return obj.status.label

    @staticmethod
    def get_slack_channel_name(obj: Incident) -> str:
        return f"#{obj.slack_channel_name}" if obj.slack_channel_name else ""

    @staticmethod
    def get_postmortem_url(obj: Incident) -> str | None:
        """Return the Confluence post-mortem page URL if it exists."""
        if hasattr(obj, "postmortem_for"):
            return obj.postmortem_for.page_url
        return None

    @staticmethod
    def get_jira_ticket_key(obj: Incident) -> str | None:
        """Return the Jira ticket key if it exists."""
        if hasattr(obj, "jira_ticket") and obj.jira_ticket:
            return obj.jira_ticket.key
        return None

    @staticmethod
    def get_jira_ticket_url(obj: Incident) -> str | None:
        """Return the Jira ticket URL if it exists."""
        if hasattr(obj, "jira_ticket") and obj.jira_ticket:
            return obj.jira_ticket.url
        return None

    def create(self, validated_data: dict[str, Any]) -> Incident:
        return Incident.objects.declare(**validated_data)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        remove_fields = get_in(kwargs, "context.remove_fields", [])
        super().__init__(*args, **kwargs)

        if remove_fields:
            # for multiple fields in a list
            for field_name in remove_fields:
                self.fields.pop(field_name)

    class Meta:
        model = Incident
        depth = 2
        fields = [
            "id",
            "title",
            "status",
            "description",
            "created_at",
            "environment",
            "incident_category",
            "priority",
            "status",
            "slack_channel_name",
            "status_page_url",
            "postmortem_url",
            "jira_ticket_key",
            "jira_ticket_url",
            "status",
            "environment_id",
            "incident_category_id",
            "priority_id",
            "created_by_email",
            "tags",
            "created_by",
            "costs",
            "metrics",
            "roles",
            "ignore",
        ]



from __future__ import annotations

import pytest
from django.conf import settings

from firefighter.incidents.factories import UserFactory
from firefighter.incidents.models.impact import (
    Impact,
    ImpactLevel,
    ImpactType,
    LevelChoices,
)
from firefighter.jira_app.models import JiraUser
from firefighter.raid.models import FeatureTeam, JiraTicket, JiraTicketImpact


@pytest.mark.django_db
class TestJiraTicket:
    def test_get_absolute_url(self):
        # Given
        user = UserFactory()
        jira_user = JiraUser.objects.create(id="jira123", user=user)

        jira_ticket = JiraTicket.objects.create(
            id=12345,
            key="TEST-123",
            summary="Test ticket",
            description="Test description",
            reporter=jira_user,
        )

        # When
        result = jira_ticket.get_absolute_url()

        # Then
        expected_url = f"{settings.RAID_JIRA_API_URL}/browse/TEST-123"
        assert result == expected_url

    def test_url_property(self):
        # Given
        user = UserFactory()
        jira_user = JiraUser.objects.create(id="jira456", user=user)

        jira_ticket = JiraTicket.objects.create(
            id=45678,
            key="TEST-456",
            summary="Test ticket 2",
            description="Test description 2",
            reporter=jira_user,
        )

        # When
        result = jira_ticket.url

        # Then
        expected_url = f"{settings.RAID_JIRA_API_URL}/browse/TEST-456"
        assert result == expected_url

    def test_url_property_equals_get_absolute_url(self):
        # Given
        user = UserFactory()
        jira_user = JiraUser.objects.create(id="jira789", user=user)

        jira_ticket = JiraTicket.objects.create(
            id=78910,
            key="TEST-789",
            summary="Test ticket 3",
            description="Test description 3",
            reporter=jira_user,
        )

        # When & Then
        assert jira_ticket.url == jira_ticket.get_absolute_url()


@pytest.mark.django_db
class TestJiraTicketImpact:
    def test_string_representation(self):
        # Given
        user = UserFactory()
        jira_user = JiraUser.objects.create(id="jira100", user=user)

        jira_ticket = JiraTicket.objects.create(
            id=10011,
            key="TEST-100",
            summary="Test ticket",
            description="Test description",
            reporter=jira_user,
        )

        # Create required objects for Impact
        impact_type = ImpactType.objects.create(name="Test Impact Type")
        impact_level = ImpactLevel.objects.create(
            name="High",
            impact_type=impact_type,
            value=LevelChoices.HIGH,
        )
        impact = Impact.objects.create(
            impact_type=impact_type,
            impact_level=impact_level,
            details="Test impact details",
        )

        jira_ticket_impact = JiraTicketImpact.objects.create(
            jira_ticket=jira_ticket,
            impact=impact,
        )

        # When
        result = str(jira_ticket_impact)

        # Then
        expected = f"{jira_ticket.key}: {impact}"
        assert result == expected


@pytest.mark.django_db
class TestFeatureTeam:
    def test_string_representation(self):
        # Given
        feature_team = FeatureTeam.objects.create(
            name="Test Team",
            jira_project_key="TST",
        )

        # When
        result = str(feature_team)

        # Then
        assert result == "Test Team"

    def test_get_team_property(self):
        # Given
        feature_team = FeatureTeam.objects.create(
            name="Backend Team",
            jira_project_key="BACK",
        )

        # When
        result = feature_team.get_team

        # Then
        expected = "Backend Team  BACK"
        assert result == expected

    def test_get_key_property(self):
        # Given
        feature_team = FeatureTeam.objects.create(
            name="Frontend Team",
            jira_project_key="FRONT",
        )

        # When
        result = feature_team.get_key

        # Then
        assert result == "FRONT"

    def test_unique_constraint(self):
        # Given - Create first team
        FeatureTeam.objects.create(
            name="Unique Team",
            jira_project_key="UNIQ",
        )

        # When & Then - Try to create duplicate should raise error
        with pytest.raises((Exception,)):
            FeatureTeam.objects.create(
                name="Unique Team",
                jira_project_key="UNIQ",
            )

    def test_jira_project_key_unique_constraint(self):
        # Given - Create first team
        FeatureTeam.objects.create(
            name="Team A",
            jira_project_key="SHARED",
        )

        # When & Then - Try to create another team with same key should raise error
        with pytest.raises((Exception,)):
            FeatureTeam.objects.create(
                name="Team B",
                jira_project_key="SHARED",
            )



# Generated manually on 2025-08-19 - Copy component relationships to incident_categories

from django.db import migrations


def copy_components_to_incident_categories(apps, schema_editor):
    """Copy all component M2M relationships to incident_categories M2M relationships"""
    Conversation = apps.get_model("slack", "Conversation")
    UserGroup = apps.get_model("slack", "UserGroup")

    # Copy conversation components to incident_categories
    conversations_updated = 0
    for conversation in Conversation.objects.prefetch_related("components").all():
        # Copy all component relationships to incident_categories (same UUIDs)
        incident_category_ids = list(conversation.components.values_list("id", flat=True))
        conversation.incident_categories.set(incident_category_ids)
        if incident_category_ids:
            conversations_updated += 1

    # Copy usergroup components to incident_categories
    usergroups_updated = 0
    for usergroup in UserGroup.objects.prefetch_related("components").all():
        incident_category_ids = list(usergroup.components.values_list("id", flat=True))
        usergroup.incident_categories.set(incident_category_ids)
        if incident_category_ids:
            usergroups_updated += 1


def reverse_copy_components_to_incident_categories(apps, schema_editor):
    """Reverse: copy incident_categories relationships back to components"""
    Conversation = apps.get_model("slack", "Conversation")
    UserGroup = apps.get_model("slack", "UserGroup")

    # Copy conversation incident_categories back to components
    for conversation in Conversation.objects.prefetch_related("incident_categories").all():
        component_ids = list(conversation.incident_categories.values_list("id", flat=True))
        conversation.components.set(component_ids)

    # Copy usergroup incident_categories back to components
    for usergroup in UserGroup.objects.prefetch_related("incident_categories").all():
        component_ids = list(usergroup.incident_categories.values_list("id", flat=True))
        usergroup.components.set(component_ids)


class Migration(migrations.Migration):

    dependencies = [
        ("slack", "0005_add_incident_categories_fields"),
        ("incidents", "0023_populate_incident_category_references"),
    ]

    operations = [
        migrations.RunPython(
            copy_components_to_incident_categories,
            reverse_copy_components_to_incident_categories,
        ),
    ]
